import numpy as np

import plotly.graph_objs as go

from dataclasses import dataclass

from plotly.subplots import make_subplots

from .processing import (
    get_colors_from_numeric_values,
    combine_sequences
)

__all__ = [
    "PlotConfig",
    "AxisConfig",
    "LayoutConfig",
    "LegendConfig",
    "config_fig",
    "plot_unfolding"
]

@dataclass
class PlotConfig:
    """General plot configuration"""
    width: int = 1000
    height: int = 800
    type: str = "png"
    font_size: int = 16
    marker_size: int = 8
    line_width: int = 3

@dataclass
class AxisConfig:
    """Axis styling configuration"""
    showgrid_x: bool = True
    showgrid_y: bool = True
    n_y_axis_ticks: int = 5
    linewidth: int = 1
    tickwidth: int = 1
    ticklen: int = 5
    gridwidth: int = 1

@dataclass
class LayoutConfig:
    """Layout and spacing configuration"""
    show_subplot_titles: bool = False
    vertical_spacing: float = 0.1

@dataclass
class LegendConfig:
    """Legend and labeling configuration"""
    color_bar_length = 0.4
    color_bar_orientation = "v"
    color_bar_x_pos = 1.05
    color_bar_y_pos = 0.5


def config_fig(fig,
               plot_width=800,
               plot_height=600,
               plot_type="png",
               plot_title_for_download="plot"):
    """
    Configure plotly figure with download options and toolbar settings.

    Parameters
    ----------
    fig : go.Figure
        Plotly figure object
    plot_width : int, default 800
        Width of the plot in pixels
    plot_height : int, default 600
        Height of the plot in pixels
    plot_type : str, default "png"
        Format for downloading the plot (e.g., "png", "jpeg")
    plot_title_for_download : str, default "plot"
        Title for the downloaded plot file

    Returns
    -------
    go.Figure
        Configured plotly figure
    """

    # Append the file extension to the title for download
    plot_title_for_download += f".{plot_type}"

    config = {
        'toImageButtonOptions': {
            'format': plot_type,
            'filename': plot_title_for_download,
            'width': plot_width,
            'height': plot_height
        },
        'displaylogo': False,
        'modeBarButtonsToRemove': [
            'sendDataToCloud',
            'hoverClosestCartesian',
            'hoverCompareCartesian',
            'lasso2d',
            'select2d'
        ]
    }

    fig.update_layout(
        width=plot_width,
        height=plot_height
    )

    fig._config = config

    return fig

def plot_unfolding(
        pychemelt_sample,
        plot_derivative = False,
        plot_config: PlotConfig = None,
        axis_config: AxisConfig = None,
        layout_config: LayoutConfig = None,
        legend_config: LegendConfig = None):

    """
    Plot the unfolding curves, including the signal and the predicted curves

    Parameters
    ----------

    pychemelt_sample:
        pychemelt.Sample object
    plot_derivative: bool
        Whether to plot the derivative of the signal
    plot_config : PlotConfig, optional
        Configuration for the overall plot
    axis_config : AxisConfig, optional
        Configuration for the axes
    layout_config : LayoutConfig, optional
        Configuration for the layout
    legend_config : LegendConfig, optional
        configuration for the legend

    """

    # Set defaults for configuration objects
    plot_config = plot_config or PlotConfig()
    axis_config = axis_config or AxisConfig()
    layout_config = layout_config or LayoutConfig()
    legend_config = legend_config or LegendConfig()

    fittings_done = pychemelt_sample.global_fit_params is not None

    # If derivative is plotted and not present, get derivative
    if plot_derivative and not hasattr(pychemelt_sample, "deriv_lst_multiple") or fittings_done and not hasattr(pychemelt_sample, "predicted_deriv_lst_multiple"):
        pychemelt_sample.estimate_derivative()

    # Extract the minimum and maximum denaturation concentration
    concs = pychemelt_sample.denaturant_concentrations

    # Adjusting scale depending on highest concentration
    scale = "M"

    if np.max(concs) < 1e-1:
        concs = concs * 1e3
        scale = "mM"
    if np.max(concs) < 1e-1:
        concs = concs * 1e3
        scale = "μM"

    min_conc = np.min(concs)
    max_conc = np.max(concs)

    colors = get_colors_from_numeric_values(concs, min_conc, max_conc)

    n_subplots = pychemelt_sample.nr_signals

    # Set number of rows: 2 if less than 8 plots, else 3
    nrows = 2 if n_subplots < 9 else 3
    nrows = min(nrows, n_subplots)  # Do not exceed the number of plots - case n equal 1

    ncols = int(np.ceil(n_subplots / nrows))

    row_arr = np.arange(1, nrows + 1)
    col_arr = np.arange(1, ncols + 1)
    # Row and column counters for subplotting
    row_col_info = combine_sequences(row_arr, col_arr)

    subplot_titles = pychemelt_sample.signal_names

    fig = make_subplots(
        rows=nrows,
        cols=ncols,
        shared_xaxes=True,
        shared_yaxes=False,
        vertical_spacing=layout_config.vertical_spacing,
        subplot_titles=subplot_titles)

    subplot_idx = 0

    ys_fit = None

    nr_den = pychemelt_sample.nr_den

    for i in range(n_subplots):

        row = row_col_info[subplot_idx][0]
        col = row_col_info[subplot_idx][1]

        if fittings_done:
            # Reduced dataset if fittings were done
            xs     = pychemelt_sample.temp_lst_expanded[i*nr_den:(i+1)*nr_den]
            if plot_derivative:
                ys_fit = pychemelt_sample.predicted_deriv_lst_multiple[i]
                ys = pychemelt_sample.deriv_lst_expanded[i * nr_den:(i + 1) * nr_den]
            else:
                ys_fit = pychemelt_sample.predicted_lst_multiple[i]
                ys = pychemelt_sample.signal_lst_expanded[i*nr_den:(i+1)*nr_den]

        else:
            # Full dataset if no fittings were done
            xs = pychemelt_sample.temp_lst_multiple[i]
            if plot_derivative:
                ys = pychemelt_sample.deriv_lst_multiple[i]
            else:
                ys = pychemelt_sample.signal_lst_multiple[i]

        for j,conc in enumerate(concs):

            color = colors[j]

            x = xs[j]
            y = ys[j]

            fig.add_trace(
                go.Scatter(
                    x=x, y=y, mode='markers',
                    marker=dict(size=plot_config.marker_size, color=color),
                    name=f'{conc:.2f} {scale}',
                    showlegend=False
                ),
                row=row, col=col
            )

            if fittings_done:

                # count np.nans in ys_fit
                ys_fit_j = ys_fit[j]

                fig.add_trace(
                    go.Scatter(
                        x=x, y=ys_fit_j, mode='lines',
                        line=dict(color='black', width=plot_config.line_width),
                        showlegend=False,
                        hoverinfo='skip',
                        hovertemplate=None
                    ),
                    row=row, col=col
                )

        subplot_idx += 1

    # Update subplot layout with white background and axis styling
    fig.update_layout(
        font_family="Roboto",
        font_color="black",
        plot_bgcolor='white',
        paper_bgcolor='white',
        legend=dict(font=dict(size=plot_config.font_size - 1))
    )

    for i in range(n_subplots):

        row = row_col_info[i][0]
        col = row_col_info[i][1]

        # Set the x-axis title only for the last row
        title_text_x = 'Temperature (°C)' if row == nrows else ''

        # Set the y-axis title only for the first column
        if plot_derivative:
            title_text_y = 'Derivative' if col == 1 else ''
        else:
            title_text_y = 'Signal' if col == 1 else ''

        fig.update_xaxes(
            title_text=title_text_x,
            showgrid=axis_config.showgrid_x,
            gridwidth=axis_config.gridwidth,
            gridcolor='lightgray',
            showline=True,
            linewidth=axis_config.linewidth,
            linecolor='black',
            zeroline=False,
            tickcolor='black',
            ticks="outside",
            tickwidth=axis_config.tickwidth,
            ticklen=axis_config.ticklen,
            title_font_size=plot_config.font_size,
            tickfont_size=plot_config.font_size,
            col = col,
            row = row
        )

        fig.update_yaxes(
            title_text=title_text_y,
            showgrid=axis_config.showgrid_y,
            gridwidth=axis_config.gridwidth,
            gridcolor='lightgray',
            showline=True,
            linewidth=axis_config.linewidth,
            linecolor='black',
            zeroline=False,
            tickcolor='black',
            ticks="outside",
            tickwidth=axis_config.tickwidth,
            ticklen=axis_config.ticklen,
            title_font_size=plot_config.font_size,
            tickfont_size=plot_config.font_size,
            nticks = axis_config.n_y_axis_ticks,
            col=col,
            row=row
        )

    # Build colorbar dict using legend_config values (orientation and x/y position)
    # Choose sensible anchors depending on orientation
    _xanchor = 'center' if legend_config.color_bar_orientation == 'h' else 'left'
    _yanchor = 'top'    if legend_config.color_bar_orientation == 'h' else 'middle'

    colorbar_dict = dict(
        title=f'[Protein] ({scale})' if pychemelt_sample.oligomeric else f'[Denaturant] ({scale})',
        tickvals=[min_conc, 0.5*(min_conc + max_conc), max_conc],
        ticktext=[f"{min_conc:.2g}", f"{(min_conc + max_conc) * 0.5:.2g}", f"{max_conc:.2g}"],
        len=legend_config.color_bar_length,
        outlinewidth=1,
        ticks='outside',
        tickfont=dict(size=plot_config.font_size - 1),
        orientation=legend_config.color_bar_orientation,
        x=legend_config.color_bar_x_pos,
        y=legend_config.color_bar_y_pos,
        xanchor=_xanchor,
        yanchor=_yanchor
    )

    fig.add_trace(
        go.Scatter(
            x=[None], y=[None], mode='markers',
            marker=dict(
                colorscale='Viridis',
                cmin=min_conc,
                cmax=max_conc,
                colorbar=colorbar_dict
            ),
            showlegend=False,
            hoverinfo='skip'
        ),
        row=1, col=1
    )

    def is_subplot_title(x):
        return x.text in ["Simulated Signal"]

    fig.update_annotations(
        selector=is_subplot_title,
        patch=dict(font=dict(size=plot_config.font_size + 10))
    )

    fig = config_fig(
        fig,
        plot_config.width,
        plot_config.height,
        plot_config.type
    )

    return fig



import numpy as np

from pychemelt.utils.svd import (
    apply_pca,
    reconstruct_spectra,
    filter_basis_spectra,
    rotate_two_basis_spectra,
    rotate_three_basis_spectra,
    recalc_explained_variance
)


# Generate test data
def gaussian(x, amp, ctr, std):
    return amp * np.exp(-(x - ctr) ** 2 / (2 * (std ** 2)))


x = np.linspace(100, 400, 300)

B1 = gaussian(x, 1, 210, 7)
B2 = gaussian(x, 1, 240, 12)
B3 = gaussian(x, 1, 270, 5)

A = np.array([
    [0.0, 1.0, 0.1],
    [0.3, 1.0, 0.4],
    [0.2, 0.5, 1.0],
    [0.9, 0.1, 0.2]
])

X_2 = (A[:, :2] @ np.vstack([B1, B2])).T + 5
X_3 = (A @ np.vstack([B1, B2, B3])).T + 5  # No need to transpose here, like in test_svd.py


def test_apply_pca():
    cum_sum_eigenvalues, principal_components, coefficients = apply_pca(X_3)

    # Check the coefficients shape
    assert len(coefficients) == X_3.shape[0]

    # Verify that the reconstructed data matches the original data
    X_reconstructed = reconstruct_spectra(principal_components, coefficients, X=X_3, pca_based=True)

    assert np.allclose(X_3, X_reconstructed)

    # We expect three significant components
    assert cum_sum_eigenvalues[1] < 100
    assert np.allclose(cum_sum_eigenvalues[2], 100, rtol=1e-5)


def test_filter_basis_spectra():
    explained_variance, basis_spectra, coefficients = apply_pca(X_2)

    _, _, k = filter_basis_spectra(explained_variance, basis_spectra, coefficients, 99.99)

    assert k == 2

    explained_variance, basis_spectra, coefficients = apply_pca(X_3)

    _, _, k = filter_basis_spectra(explained_variance, basis_spectra, coefficients, 99.99)

    assert k == 3


def test_rotate_spectra():
    # Rotate two basis spectra
    explained_variance, basis_spectra, coefficients = apply_pca(X_2)

    basis_spectra_new, _, _ = filter_basis_spectra(explained_variance, basis_spectra, coefficients, 100)

    basis_spectra_rot, coefficients_rot = rotate_two_basis_spectra(X_2, basis_spectra_new, pca_based=True)

    # Verify that the reconstructed data still matches the original data
    X_reconstructed = reconstruct_spectra(basis_spectra_rot, coefficients_rot, X=X_2, pca_based=True)

    assert np.allclose(X_2, X_reconstructed)

    # Rotate three basis spectra
    explained_variance, basis_spectra, coefficients = apply_pca(X_3)

    basis_spectra_new, _, _ = filter_basis_spectra(explained_variance, basis_spectra, coefficients, 100)

    basis_spectra_rot, coefficients_rot = rotate_three_basis_spectra(X_3, basis_spectra_new, pca_based=True)

    # Verify that the reconstructed data still matches the original data
    X_reconstructed = reconstruct_spectra(basis_spectra_rot, coefficients_rot, X=X_3, pca_based=True)

    assert np.allclose(X_3, X_reconstructed)


def test_variance_change_after_rotation():
    # Rotate two basis spectra
    explained_variance, basis_spectra, coefficients = apply_pca(X_2)

    basis_spectra_new, _, _ = filter_basis_spectra(explained_variance, basis_spectra, coefficients, 100)

    basis_spectra_rot, coefficients_rot = rotate_two_basis_spectra(X_2, basis_spectra_new, pca_based=True)

    # Recalculate explained variance
    explained_variance_rot = recalc_explained_variance(basis_spectra_rot, coefficients_rot, X_2, pca_based=True)

    assert explained_variance_rot[0] < explained_variance[0]
    assert np.allclose(explained_variance_rot[1], explained_variance[1], rtol=1e-5)


import pytest
import numpy as np

from pychemelt.utils.fractions import (
    fn_two_state_monomer,
    fu_two_state_dimer,
    fu_two_state_trimer,
    fu_two_state_tetramer
)


class TestMonomerFraction:
    """Tests for fn_two_state_monomer function"""

    def test_monomer_fully_folded(self):
        """When K=0 (strongly favors native), fraction folded should be 1"""
        K = 0.0
        fn = fn_two_state_monomer(K)
        assert fn == pytest.approx(1.0)

    def test_monomer_fully_unfolded(self):
        """When K is very large (strongly favors unfolded), fraction folded should approach 0"""
        K = 1e10
        fn = fn_two_state_monomer(K)
        assert fn == pytest.approx(0.0, abs=1e-9)

    def test_monomer_equal_populations(self):
        """When K=1 (equal populations), fraction folded should be 0.5"""
        K = 1.0
        fn = fn_two_state_monomer(K)
        assert fn == pytest.approx(0.5)

    def test_monomer_typical_values(self):
        """Test with typical equilibrium constant values"""
        # K=0.1 should give fn ≈ 0.909
        assert fn_two_state_monomer(0.1) == pytest.approx(0.909, abs=1e-3)

        # K=10 should give fn ≈ 0.091
        assert fn_two_state_monomer(10.0) == pytest.approx(0.091, abs=1e-3)

        # K=0.25 should give fn = 0.8
        assert fn_two_state_monomer(0.25) == pytest.approx(0.8)

    def test_monomer_array_input(self):
        """Test that function works with numpy arrays"""
        K = np.array([0.0, 0.5, 1.0, 2.0, 10.0])
        fn = fn_two_state_monomer(K)

        expected = np.array([1.0, 2/3, 0.5, 1/3, 1/11])
        np.testing.assert_allclose(fn, expected, rtol=1e-10)


class TestDimerFraction:
    """Tests for fu_two_state_dimer function (N2 <-> 2U)"""

    def test_dimer_fully_folded(self):
        """When K=0, all dimer should be folded, fu=0"""
        K = 1e-30
        C = 1e-6  # 1 µM
        fu = fu_two_state_dimer(K, C)
        assert fu == pytest.approx(0.0, abs=1e-10)

    def test_dimer_fully_unfolded(self):
        """When K is very large, dimer should be fully unfolded, fu→1"""
        K = 1e10
        C = 1e-6
        fu = fu_two_state_dimer(K, C)
        assert fu == pytest.approx(1.0, abs=1e-6)

    def test_dimer_typical_values(self):
        """Test with typical K and concentration values"""
        # Test case 1: Moderate K and concentration
        K = 1e-2
        C = 1
        fu = fu_two_state_dimer(K, C)
        # For dimer: fu² = K/(4C(1-fu)) -> should be partially unfolded
        assert 0.0 < fu < 1.0


    def test_dimer_concentration_dependence(self):
        """Test that higher concentration shifts equilibrium toward folded state"""
        K = 1e-6
        C_low = 1e-7
        C_high = 1e-5

        fu_low = fu_two_state_dimer(K, C_low)
        fu_high = fu_two_state_dimer(K, C_high)

        # Higher concentration should favor folded state (lower fu)
        assert fu_low > fu_high

    def test_dimer_array_input(self):
        """Test that function works with numpy arrays"""
        K = 1e-6
        C = np.array([1e-7, 1e-6, 1e-5])
        fu = fu_two_state_dimer(K, C)

        assert len(fu) == len(C)
        assert np.all(fu >= 0) and np.all(fu <= 1)


class TestTrimerFraction:
    """Tests for fu_two_state_trimer function (N3 <-> 3U)"""

    def test_trimer_fully_folded(self):
        """When K=0, all trimer should be folded, fu=0"""
        K = 0.0
        C = 1e-6
        fu = fu_two_state_trimer(K, C)
        assert fu == pytest.approx(0.0, abs=1e-10)

    def test_trimer_fully_unfolded(self):
        """When K is very large, trimer should be fully unfolded, fu→1"""
        K = 1e15
        C = 1e-6
        fu = fu_two_state_trimer(K, C)
        assert fu == pytest.approx(1.0, abs=1e-3)

    def test_trimer_typical_values(self):
        """Test with typical K and concentration values"""
        K = 1e-12
        C = 1e-6
        fu = fu_two_state_trimer(K, C)

        # Should be partially unfolded
        assert 0.0 < fu < 1.0


    def test_trimer_concentration_dependence(self):
        """Test that higher concentration shifts equilibrium toward folded state"""
        K = 1e-12
        C_low = 1e-7
        C_high = 1e-5

        fu_low = fu_two_state_trimer(K, C_low)
        fu_high = fu_two_state_trimer(K, C_high)

        # Higher concentration should favor folded state (lower fu)
        assert fu_low > fu_high

    def test_trimer_array_input(self):
        """Test that function works with numpy arrays"""
        K = 1e-12
        C = np.array([1e-7, 1e-6, 1e-5])
        fu = fu_two_state_trimer(K, C)

        assert len(fu) == len(C)
        assert np.all(fu >= 0) and np.all(fu <= 1)


class TestTetramerFraction:
    """Tests for fu_two_state_tetramer function (N4 <-> 4U)"""

    def test_tetramer_fully_folded(self):
        """When K=0, all tetramer should be folded, fu=0"""
        K = 0.0
        C = 1e-6
        fu = fu_two_state_tetramer(K, C)
        assert fu == pytest.approx(0.0, abs=1e-10)

    def test_tetramer_fully_unfolded(self):
        """When K is very large, tetramer should be fully unfolded, fu→1"""
        K = 1e20
        C = 1e-6
        fu = fu_two_state_tetramer(K, C)
        assert fu == pytest.approx(1.0, abs=1e-2)

    def test_tetramer_typical_values(self):
        """Test with typical K and concentration values"""
        K = 1e-18
        C = 1e-6
        fu = fu_two_state_tetramer(K, C)

        # Should be partially unfolded
        assert 0.0 < fu < 1.0

        # Verify the equilibrium relationship: K = 256*C³*fu⁴/(1-fu)⁴
        fn = 1 - fu
        if fn > 0:
            K_calc = 256 * C**3 * fu**4 / (fn**4)
            assert K_calc == pytest.approx(K, rel=1e-2)

    def test_tetramer_concentration_dependence(self):
        """Test that higher concentration shifts equilibrium toward folded state"""
        K = 1e-18
        C_low = 1e-7
        C_high = 1e-5

        fu_low = fu_two_state_tetramer(K, C_low)
        fu_high = fu_two_state_tetramer(K, C_high)

        # Higher concentration should favor folded state (lower fu)
        assert fu_low > fu_high

    def test_tetramer_array_input(self):
        """Test that function works with numpy arrays"""
        K = 1e-18
        C = np.array([1e-7, 1e-6, 1e-5])
        fu = fu_two_state_tetramer(K, C)

        assert len(fu) == len(C)
        # All values should be valid fractions (allowing for numerical edge cases)
        assert np.all((fu >= 0) & (fu <= 1.01))

    def test_tetramer_numerical_stability(self):
        """Test that tetramer calculation handles edge cases properly"""
        # Very small K
        fu = fu_two_state_tetramer(1e-50, 1e-6)
        assert fu == pytest.approx(0.0, abs=1e-5)

        # Moderate values that might cause numerical issues
        K = np.array([1e-20, 1e-18, 1e-16])
        C = 1e-6
        fu = fu_two_state_tetramer(K, C)

        # Should all be valid fractions
        assert np.all((fu >= 0) & (fu <= 1.01))
        # Should be monotonically increasing with K
        assert np.all(np.diff(fu) > 0)






"""
Module for thermal unfolding simulation and testing with given concentration and model parameters.

This module includes functionalities for simulating thermal unfolding using the ThermalOligomer class,
applying two-state models for signal mapping, baseline estimation, and curve fitting. Various test
functions are provided to validate the implementation, including parameter fitting for monomer
models and scaling behaviors.

Classes:
    ThermalOligomer: Represents the system for thermal unfolding simulations and parameter fitting.

Notes
-----
- The tests rely on seeded random number generators for reproducible results.
- Thermal unfolding behaviors can accommodate different baselines (linear, exponential).
- The module utilizes the `pytest` library for test development.

"""

import numpy as np

import pytest

from pychemelt.thermal_oligomer import ThermalOligomer

from pychemelt.utils.math import linear_baseline, exponential_baseline

from pychemelt.utils.signals import (
    map_two_state_model_to_signal_fx
)


# Centralized test constants
RNG_SEED = 2
TEMP_START = 20.0
TEMP_STOP = 90.0
N_TEMPS = 150
CONCS = np.arange(10, 80, 10)*1e-6

# Model / ground-truth parameters
DHm_VAL = 100
Tm_VAL = 70
CP0_VAL = 1.0

DHm_INCREASE = 50

INTERCEPT_N = 24
SLOPE_N = -0.27
C_N_VAL = 0
INTERCEPT_U = -4
SLOPE_U = 80.5
EXPONENT_U = 0.0224
C_U_VAL = 0

rng = np.random.default_rng(RNG_SEED)

def_params = {
    'dHm': DHm_VAL,
    'Tm': Tm_VAL+273.15,
    'Cp': CP0_VAL,
    'p1_N': C_N_VAL,
    'p2_N': INTERCEPT_N,
    'p3_N': SLOPE_N,
    'p4_N': 0,
    'p1_U': C_U_VAL,
    'p2_U': INTERCEPT_U,
    'p3_U': SLOPE_U,
    'p4_U': EXPONENT_U,
    'baseline_N_fx':linear_baseline,
    'baseline_U_fx':exponential_baseline

}

concs = CONCS

def aux_create_pychem_sim(params,concs, model, normalise=False):

    signal_fx = map_two_state_model_to_signal_fx(model)

    # Calculate signal range for proper y-axis scaling
    temp_range  = np.linspace(TEMP_START, TEMP_STOP, N_TEMPS)
    temp_range_K = temp_range + 273.15

    signal_list = []
    temp_list   = []

    # Use a seeded Generator for reproducible noise in tests
    rng = np.random.default_rng(2)

    for C in concs:

        y = signal_fx(temp_range_K, C, **params)

        # Add gaussian error to signal
        y += rng.normal(0, 0.002*1e-3, len(y)) # Small error (seeded)

        signal_list.append(y)
        temp_list.append(temp_range)

    pychem_sim = ThermalOligomer()

    pychem_sim.set_model(model)

    pychem_sim.signal_dic['Fluo'] = signal_list
    pychem_sim.temp_dic['Fluo']   = [temp_range for _ in range(len(concs))]

    pychem_sim.conditions = concs

    pychem_sim.global_min_temp = np.min(temp_range)
    pychem_sim.global_max_temp = np.max(temp_range)

    pychem_sim.set_concentrations()

    pychem_sim.set_signal(['Fluo'])

    pychem_sim.select_conditions(normalise_to_global_max=normalise)
    pychem_sim.expand_multiple_signal()



    pychem_sim.estimate_baseline_parameters(
        native_baseline_type='linear',
        unfolded_baseline_type='exponential'
    )

    pychem_sim.n_residues = 80  # only for cp initial guess
    pychem_sim.guess_Cp()

    return pychem_sim


sample = ThermalOligomer()
sample.read_multiple_files('./test_files/nDSFdemoFile.xlsx')


def test_set_model():

    sample.set_model("Monomer")
    assert sample.model == "Monomer"

    sample.set_model("Dimer")
    assert sample.model == "Dimer"

    sample.set_model("dimer")
    assert sample.model == "Dimer"

    pytest.raises(ValueError, sample.set_model, "test_false")


def test_set_concentrations():

    sample.set_concentrations()

    assert np.min(sample.oligomer_concentrations_pre) == 0
    assert np.max(sample.oligomer_concentrations_pre) == 8.24

def test_select_conditions():
    sample.set_signal(['350nm'])

    # Select without scaling
    sample.select_conditions(
        [False for _ in range(24)] + [True for _ in range(8)] + [False for _ in range(16)],
        normalise_to_global_max=False
    )

    assert len(sample.signal_lst_multiple) == 1
    assert len(sample.signal_lst_multiple[0]) == 8
    assert np.max(sample.signal_lst_multiple[0]) != 1.0

    # Select with scaling
    sample.select_conditions(
        [False for _ in range(24)] + [True for _ in range(8)] + [False for _ in range(16)],
        normalise_to_global_max=True
    )

    assert len(sample.signal_lst_multiple) == 1
    assert len(sample.signal_lst_multiple[0]) == 8
    assert np.max(sample.signal_lst_multiple[0]) == 100

def test_guess_Cp():

    sample.n_residues = 150
    sample.guess_Cp()

    assert sample.Cp0 == pytest.approx(2.0933, rel=1e-2)

    sample.n_residues = 0
    pytest.raises(ValueError, sample.guess_Cp)

    sample.n_residues = 5

    sample.guess_Cp()
    assert sample.Cp0 == pytest.approx(0, rel=1e-2)


def test_fit_thermal_unfolding_global_global_global_failure():

    sample = ThermalOligomer()
    sample.read_multiple_files('./test_files/nDSFdemoFile.xlsx')

    pytest.raises(ValueError, sample.fit_thermal_unfolding_global_global_global)

def test_fit_thermal_unfolding_global_global_global_scaling():
    model = "Monomer"
    rng = np.random.default_rng(RNG_SEED)

    #Using concentrations close to each other in order to trigger non-scaling
    scale_concs = [0.999999999999999, 1.00000000000000000001]

    temp_range = np.linspace(20, 90, 100)
    temp_range_K = temp_range + 273.15

    signal_list = []
    temp_list = []

    signal_fx = map_two_state_model_to_signal_fx(model)

    for i, C in enumerate(scale_concs):
        y = signal_fx(temp_range_K, C, **def_params)

        # Concentration dependent scaling
        y = y * C

        # Add gaussian error to simulated signal
        y += rng.normal(0, 0.02, len(y))

        # Add error to the initial signal to model variance across positions
        y *= rng.uniform(0.9, 1.1)

        signal_list.append(y)
        temp_list.append(temp_range)

    pychem_sim = ThermalOligomer()

    pychem_sim.signal_dic['Simulated signal'] = signal_list
    pychem_sim.temp_dic['Simulated signal'] = [temp_range for _ in range(len(scale_concs))]

    pychem_sim.set_model(model)
    pychem_sim.conditions = scale_concs

    pychem_sim.global_min_temp = np.min(temp_range)
    pychem_sim.global_max_temp = np.max(temp_range)

    pychem_sim.set_concentrations()

    pychem_sim.set_signal('Simulated signal')

    pychem_sim.select_conditions(normalise_to_global_max=True)
    pychem_sim.expand_multiple_signal()

    pychem_sim.estimate_baseline_parameters(
        native_baseline_type='linear',
        unfolded_baseline_type='exponential'
    )

    pychem_sim.n_residues = 130  # only for cp initial guess
    pychem_sim.guess_Cp()

    pychem_sim.fit_thermal_unfolding_global()
    pychem_sim.fit_thermal_unfolding_global_global()
    pychem_sim.fit_thermal_unfolding_global_global_global(model_scale_factor=True)

    assert pychem_sim.params_df is not None

# Testing Monomer model

monomer_sim = aux_create_pychem_sim(def_params, concs, "Monomer")

def test_fit_thermal_unfolding_global_monomer():

    # local slopes and baselines
    expected = [Tm_VAL, DHm_VAL, CP0_VAL]

    monomer_sim.fit_thermal_unfolding_global()

    np.testing.assert_allclose(monomer_sim.params_df.iloc[:3,1], expected, rtol=0.2, atol=1.5)

    # fixed Tm limits

    monomer_sim.fit_thermal_unfolding_global(tm_limits=[Tm_VAL-12, Tm_VAL+20])

    np.testing.assert_allclose(monomer_sim.params_df.iloc[:3, 1], expected, rtol=0.2, atol=1.5)

    # fixed dh limits

    monomer_sim.fit_thermal_unfolding_global(dh_limits=[10, 500])

    np.testing.assert_allclose(monomer_sim.params_df.iloc[:3, 1], expected, rtol=0.2, atol=1.5)

    # fixed cp limits

    monomer_sim.fit_thermal_unfolding_global(cp_limits=[0.1, 5])

    np.testing.assert_allclose(monomer_sim.params_df.iloc[:3, 1], expected, rtol=0.2, atol=1.5)

    # fixed cp

    expected = [Tm_VAL, DHm_VAL]

    monomer_sim.fit_thermal_unfolding_global(cp_value=CP0_VAL)

    np.testing.assert_allclose(monomer_sim.params_df.iloc[:2, 1], expected, rtol=0.2, atol=1.5)

def test_fit_thermal_unfolding_global_global_monomer():
    expected = [Tm_VAL, DHm_VAL, CP0_VAL, SLOPE_N, SLOPE_U, EXPONENT_U]

    monomer_sim.fit_thermal_unfolding_global()

    monomer_sim.fit_thermal_unfolding_global_global()

    np.testing.assert_allclose(monomer_sim.params_df.iloc[[0, 1, 2, 17, 18, 19], 1], expected, rtol=0.2, atol=1.5)

def test_fit_thermal_unfolding_global_global_global_monomer():
    expected = [Tm_VAL, DHm_VAL, CP0_VAL, INTERCEPT_N, INTERCEPT_U, SLOPE_N, SLOPE_U, EXPONENT_U]

    monomer_sim.fit_thermal_unfolding_global_global_global(model_scale_factor=True)

    np.testing.assert_allclose(monomer_sim.params_df.iloc[[0, 1, 2, 3, 4, 5, 6, 9], 1], expected, rtol=0.2, atol=1.5)

# Testing Dimer model
def_params['dHm'] = def_params['dHm'] + DHm_INCREASE

dimer_sim = aux_create_pychem_sim(def_params, concs, "Dimer")

def test_fit_thermal_unfolding_global_dimer():
    # local slopes and baselines
    expected = [Tm_VAL, DHm_VAL + DHm_INCREASE, CP0_VAL]

    dimer_sim.fit_thermal_unfolding_global()

    np.testing.assert_allclose(dimer_sim.params_df.iloc[:3, 1], expected, rtol=0.2, atol=1.5)

    # fixed Tm limits

    dimer_sim.fit_thermal_unfolding_global(tm_limits=[Tm_VAL-12, Tm_VAL+20])

    np.testing.assert_allclose(dimer_sim.params_df.iloc[:3, 1], expected, rtol=0.2, atol=1.5)

    # fixed dh limits

    dimer_sim.fit_thermal_unfolding_global(dh_limits=[10, 500])

    np.testing.assert_allclose(dimer_sim.params_df.iloc[:3, 1], expected, rtol=0.2, atol=1.5)

    # fixed cp limits

    dimer_sim.fit_thermal_unfolding_global(cp_limits=[0.1, 5])

    np.testing.assert_allclose(dimer_sim.params_df.iloc[:3, 1], expected, rtol=0.2, atol=1.5)

    # fixed cp

    expected = [Tm_VAL, DHm_VAL + DHm_INCREASE]

    dimer_sim.fit_thermal_unfolding_global(cp_value=CP0_VAL)

    np.testing.assert_allclose(dimer_sim.params_df.iloc[:2, 1], expected, rtol=0.2, atol=1.5)


def test_fit_thermal_unfolding_global_global_dimer():
    expected = [Tm_VAL, DHm_VAL + DHm_INCREASE, CP0_VAL, SLOPE_N, SLOPE_U, EXPONENT_U]

    dimer_sim.fit_thermal_unfolding_global()

    dimer_sim.fit_thermal_unfolding_global_global()

    np.testing.assert_allclose(dimer_sim.params_df.iloc[[0, 1, 2, 17, 18, 19], 1], expected, rtol=0.2, atol=1.5)

def test_fit_thermal_unfolding_global_global_global_dimer():
    expected = [Tm_VAL, DHm_VAL + DHm_INCREASE, CP0_VAL, INTERCEPT_N, INTERCEPT_U, SLOPE_N, SLOPE_U, EXPONENT_U]

    dimer_sim.fit_thermal_unfolding_global_global_global(model_scale_factor=True)

    np.testing.assert_allclose(dimer_sim.params_df.iloc[[0, 1, 2, 3, 4, 5, 6, 9], 1], expected, rtol=0.2, atol=1.5)


# Testing Trimer model
def_params['dHm'] = def_params['dHm'] + DHm_INCREASE

trimer_sim = aux_create_pychem_sim(def_params, concs, "Trimer")

def test_fit_thermal_unfolding_global_trimer():
    # local slopes and baselines
    expected = [Tm_VAL, DHm_VAL + 2*DHm_INCREASE, CP0_VAL]

    trimer_sim.fit_thermal_unfolding_global()

    np.testing.assert_allclose(trimer_sim.params_df.iloc[:3, 1], expected, rtol=0.2, atol=1.5)

    # fixed Tm limits

    trimer_sim.fit_thermal_unfolding_global(tm_limits=[Tm_VAL-12, Tm_VAL+20])

    np.testing.assert_allclose(trimer_sim.params_df.iloc[:3, 1], expected, rtol=0.2, atol=1.5)

    # fixed dh limits

    trimer_sim.fit_thermal_unfolding_global(dh_limits=[10, 500])

    np.testing.assert_allclose(trimer_sim.params_df.iloc[:3, 1], expected, rtol=0.2, atol=1.5)

    # fixed cp limits

    trimer_sim.fit_thermal_unfolding_global(cp_limits=[0.1, 5])

    np.testing.assert_allclose(trimer_sim.params_df.iloc[:3, 1], expected, rtol=0.2, atol=1.5)

    # fixed cp

    expected = [Tm_VAL, DHm_VAL + 2*DHm_INCREASE]

    trimer_sim.fit_thermal_unfolding_global(cp_value=CP0_VAL)

    np.testing.assert_allclose(trimer_sim.params_df.iloc[:2, 1], expected, rtol=0.2, atol=1.5)


def test_fit_thermal_unfolding_global_global_trimer():
    expected = [Tm_VAL, DHm_VAL + 2*DHm_INCREASE, CP0_VAL, SLOPE_N, SLOPE_U, EXPONENT_U]

    trimer_sim.fit_thermal_unfolding_global()

    trimer_sim.fit_thermal_unfolding_global_global()

    np.testing.assert_allclose(trimer_sim.params_df.iloc[[0, 1, 2, 17, 18, 19], 1], expected, rtol=0.2, atol=1.5)

def test_fit_thermal_unfolding_global_global_global_trimer():
    expected = [Tm_VAL, DHm_VAL + 2*DHm_INCREASE, CP0_VAL, INTERCEPT_N, INTERCEPT_U, SLOPE_N, SLOPE_U, EXPONENT_U]

    trimer_sim.fit_thermal_unfolding_global_global_global(model_scale_factor=True)

    np.testing.assert_allclose(trimer_sim.params_df.iloc[[0, 1, 2, 3, 4, 5, 6, 9], 1], expected, rtol=0.2, atol=1.5)


# Testing Tetramer model
def_params['dHm'] = def_params['dHm'] + DHm_INCREASE

tetramer_sim = aux_create_pychem_sim(def_params, concs, "Tetramer")

def test_fit_thermal_unfolding_global_tetramer():
    # local slopes and baselines
    expected = [Tm_VAL, DHm_VAL + 3*DHm_INCREASE, CP0_VAL]

    tetramer_sim.fit_thermal_unfolding_global()

    np.testing.assert_allclose(tetramer_sim.params_df.iloc[:3, 1], expected, rtol=0.2, atol=1.5)

    # fixed Tm limits

    tetramer_sim.fit_thermal_unfolding_global(tm_limits=[Tm_VAL - 12, Tm_VAL + 20])

    np.testing.assert_allclose(tetramer_sim.params_df.iloc[:3, 1], expected, rtol=0.2, atol=1.5)

    # fixed dh limits

    tetramer_sim.fit_thermal_unfolding_global(dh_limits=[10, 500])

    np.testing.assert_allclose(tetramer_sim.params_df.iloc[:3, 1], expected, rtol=0.2, atol=1.5)

    # fixed cp limits

    tetramer_sim.fit_thermal_unfolding_global(cp_limits=[0.1, 5])

    np.testing.assert_allclose(tetramer_sim.params_df.iloc[:3, 1], expected, rtol=0.2, atol=1.5)

    # fixed cp

    expected = [Tm_VAL, DHm_VAL + 3*DHm_INCREASE]

    tetramer_sim.fit_thermal_unfolding_global(cp_value=CP0_VAL)

    np.testing.assert_allclose(tetramer_sim.params_df.iloc[:2, 1], expected, rtol=0.2, atol=1.5)


def test_fit_thermal_unfolding_global_global_tetramer():
    expected = [Tm_VAL, DHm_VAL + 3*DHm_INCREASE, CP0_VAL, SLOPE_N, SLOPE_U, EXPONENT_U]

    tetramer_sim.fit_thermal_unfolding_global()

    tetramer_sim.fit_thermal_unfolding_global_global()

    np.testing.assert_allclose(tetramer_sim.params_df.iloc[[0, 1, 2, 17, 18, 19], 1], expected, rtol=0.2, atol=1.5)

def test_fit_thermal_unfolding_global_global_global_tetramer():
    expected = [Tm_VAL, DHm_VAL + 3*DHm_INCREASE, CP0_VAL, INTERCEPT_N, INTERCEPT_U, SLOPE_N, SLOPE_U, EXPONENT_U]

    tetramer_sim.fit_thermal_unfolding_global_global_global(model_scale_factor=True)

    np.testing.assert_allclose(tetramer_sim.params_df.iloc[[0, 1, 2, 3, 4, 5, 6, 9], 1], expected, rtol=0.2, atol=1.5)

# generating failing fit

def_params['dHm'] = 120
trimer_sim_fail = aux_create_pychem_sim(def_params, concs, "Trimer")

def test_fit_thermal_unfolding_global_warning():

    with pytest.warns(UserWarning):
        trimer_sim_fail.fit_thermal_unfolding_global()

# generating failing fit

def_params['dHm'] = 120
trimer_sim_fail_normalise = aux_create_pychem_sim(def_params, concs, "Trimer", normalise=True)

def test_fit_thermal_unfolding_global_warning_normalise():

    with pytest.warns(UserWarning):
        trimer_sim_fail_normalise.fit_thermal_unfolding_global()



def test_signal_to_df():

    signal_type_options = ['raw','derivative']

    for signal_type in signal_type_options:

        df = monomer_sim.signal_to_df(signal_type=signal_type, scaled=False)

        assert len(df) == len(concs) * N_TEMPS

    signal_type_options = ['raw','fitted']

    for signal_type in signal_type_options:

        df = monomer_sim.signal_to_df(signal_type=signal_type, scaled=True)

        assert len(df) == len(concs) * N_TEMPS
        assert np.max(df['Signal']) <= 100


        monomer_sim.max_points = 200

        df = monomer_sim.signal_to_df(signal_type=signal_type, scaled=False)

        assert df is not None


"""
Tests to ensure that the main functionalities of the pychemelt Sample class work as expected.
The order of the tests is important, as some functions depend on the previous ones.
"""
import numpy as np

from pychemelt import Monomer as Sample
from pychemelt.utils.signals import signal_two_state_tc_unfolding

from pychemelt.utils.math import quadratic_baseline

def_params = {
    'DHm': 100,
    'Tm': 60 + 273.15,
    'Cp0': 1.6,
    'm0': 2.6,
    'm1': 0,
    'p1_N': -0.1,
    'p2_N': 1.5,
    'p3_N': -0.015,  # Negative temperature dependence for native state
    'p4_N': 0.0001,
    'p1_U': -0.005,
    'p2_U': 2.5,
    'p3_U': -0.025,  # Negative temperature dependence for unfolded state
    'p4_U': 0.0002,
    'baseline_N_fx':quadratic_baseline,
    'baseline_U_fx':quadratic_baseline
}

def_concs = [1e-8,1,1.5,2,2.6,3,4,5]

scalings_factors = np.array([1,0.95,1,1.1,1,1,1,0.96])

def aux_create_pychem_sim(params,concs):

    # Calculate signal range for proper y-axis scaling
    temp_range = np.linspace(20, 80, 60)
    temp_range_K = temp_range + 273.15
    signal_list = []
    temp_list   = []

    for i,D in enumerate(concs):

        y = signal_two_state_tc_unfolding(temp_range_K, D, **params)

        rng = np.random.default_rng(2)

        # Add gaussian error to signal
        y += rng.normal(0, 0.0005, len(y)) # Small error

        # Add gaussian error to PROTEIN concentration
        y *= scalings_factors[i]

        signal_list.append(y)
        temp_list.append(temp_range)

    pychem_sim = Sample()

    pychem_sim.signal_dic['Fluo'] = signal_list
    pychem_sim.temp_dic['Fluo']   = [temp_range for _ in range(len(concs))]

    pychem_sim.conditions = concs

    pychem_sim.global_min_temp = np.min(temp_range)
    pychem_sim.global_max_temp = np.max(temp_range)

    pychem_sim.set_denaturant_concentrations()

    pychem_sim.set_signal('Fluo')

    pychem_sim.select_conditions(normalise_to_global_max=False)
    pychem_sim.expand_multiple_signal()

    return pychem_sim

def test_estimate_baseline_parameters():

    params = def_params.copy()

    # Set fluorescence dependence on temperature and denaturant concentration to zero
    params['p1_N'] = 0
    params['p1_U'] = 0
    params['p3_N'] = 0
    params['p3_U'] = 0
    params['p4_N'] = 0
    params['p4_N'] = 0

    pychem_sim = aux_create_pychem_sim(params,def_concs)

    pychem_sim.estimate_baseline_parameters(
        native_baseline_type='constant',
        unfolded_baseline_type='constant'
    )

    np.testing.assert_allclose(
        pychem_sim.first_param_Ns_per_signal[0][0],
        1.5,
        rtol=0.01,
        atol=0)

    # Reset fittings results
    sample.reset_fittings_results()
    assert len(sample.first_param_Ns_per_signal) == 0


    # ------------ #
    params = def_params.copy()

    params['p1_N'] = 0
    params['p1_U'] = 0
    params['p4_N'] = 0
    params['p4_U'] = 0

    pychem_sim = aux_create_pychem_sim(params,def_concs)

    pychem_sim.estimate_baseline_parameters(
        native_baseline_type='linear',
        unfolded_baseline_type='linear'
    )

    np.testing.assert_allclose(pychem_sim.second_param_Ns_per_signal[0][0], params['p3_N'], rtol=0.1, atol=0)
    np.testing.assert_allclose(pychem_sim.second_param_Us_per_signal[0][-1], params['p3_U'], rtol=0.1, atol=0)

    # ------------ #
    params = def_params.copy()
    params['p1_N'] = 0
    params['p1_U'] = 0

    pychem_sim = aux_create_pychem_sim(params,def_concs)

    pychem_sim.estimate_baseline_parameters(
        native_baseline_type='quadratic',
        unfolded_baseline_type='quadratic',
        window_range_native=20,
        window_range_unfolded=20
        )

    np.testing.assert_allclose(pychem_sim.third_param_Ns_per_signal[0][0], params['p4_N'], rtol=0.1, atol=0)

# --------- #  Create global pychem_sim object for the rest of tests  # --------- #
sample = aux_create_pychem_sim(def_params,def_concs)
sample.estimate_derivative()
sample.guess_Tm()
sample.n_residues = 130

def test_fit_thermal_unfolding_local():

    sample.estimate_baseline_parameters(
        native_baseline_type='quadratic',
        unfolded_baseline_type='quadratic',
        window_range_native=16,
        window_range_unfolded=16
    )
    sample.fit_thermal_unfolding_local()

    np.testing.assert_allclose(sample.Tms_multiple[0][0],60.2,rtol=0.05)

def test_guess_Cp():

    sample.guess_Cp()

    np.testing.assert_allclose(sample.Cp0,1.7,rtol=0.1)

def test_guess_initial_parameters():

    sample.guess_initial_parameters(
        native_baseline_type='quadratic',
        unfolded_baseline_type='quadratic',
        window_range_native=16,
        window_range_unfolded=16
    )

    np.testing.assert_allclose(sample.Cp0,1.7,rtol=0.2)

def test_fit_thermal_unfolding_global():

    sample.fit_thermal_unfolding_global()

    assert sample.params_df is not None

    assert sample.params_df.shape[0] == 52

    expected = [60.2, 100, 1.7, 2.6]
    actual   = sample.params_df.iloc[:4,1]

    np.testing.assert_allclose(actual,expected,rtol=0.1)

    args_dic = {
        'fit_m_dep': True,
        'dh_limits': [50,200],
        'tm_limits': [40,80],
        'cp_limits': [0.5,4]
    }

    for key,val in args_dic.items():

        sample.fit_thermal_unfolding_global(**{key:val})
        actual = sample.params_df.iloc[:4,1]
        np.testing.assert_allclose(actual,expected,rtol=0.1)

    # -- Fit with fixed Cp -- #
    sample.fit_thermal_unfolding_global(cp_value=1.7)

    expected = [60.2, 100, 2.6]
    actual = sample.params_df.iloc[:3,1]

    np.testing.assert_allclose(actual,expected,rtol=0.1)

def test_fit_thermal_unfolding_global_global():

    sample.global_fit_done = False # Force re-fitting

    sample.fit_thermal_unfolding_global_global()

    expected = [60.2, 100, 1.7, 2.6]
    actual = sample.params_df.iloc[:4,1]

    np.testing.assert_allclose(actual,expected,rtol=0.1)

def test_fit_thermal_unfolding_global_global_global():

    sample.fit_thermal_unfolding_global() # Needs to be done firsts
    sample.global_global_fit_done = False # Force re-fitting clause
    sample.fit_thermal_unfolding_global_global_global(model_scale_factor=True)

    expected = [60.2, 100, 1.7, 2.6]
    actual = sample.params_df.iloc[:4,1]

    np.testing.assert_allclose(actual,expected,rtol=0.1)


def test_signal_to_df():

    signal_type_options = ['raw','derivative']

    for signal_type in signal_type_options:

        df = sample.signal_to_df(signal_type=signal_type, scaled=False)

        assert len(df) == 480

    signal_type_options = ['raw','fitted']

    for signal_type in signal_type_options:

        df = sample.signal_to_df(signal_type=signal_type, scaled=True)

        assert len(df) == 480
        assert np.max(df['Signal']) <= 100



"""
This module contains helper functions to fit unfolding data
Author: Osvaldo Burastero
"""

import numpy as np
from scipy.optimize     import curve_fit
from scipy.optimize     import least_squares
from numpy.linalg       import pinv


from .math import get_rss, temperature_to_kelvin, temperature_to_celsius

__all__ = [
    "fit_line_robust",
    "fit_quadratic_robust",
    "fit_exponential_robust",
    "fit_thermal_unfolding",
    "fit_tc_unfolding_single_slopes",
    "fit_tc_unfolding_shared_slopes_many_signals",
    "fit_tc_unfolding_many_signals",
    "fit_oligomer_unfolding_single_slopes",
    "fit_oligomer_unfolding_shared_slopes_many_signals",
    "fit_oligomer_unfolding_many_signals",
]

def baseline_fx_name_to_req_params(baseline_fx_name):
    """
    Determine which baseline parameters are required based on the function name.

    Parameters
    ----------
    baseline_fx_name : str or function object
        baseline type to discern the number of parameters needed

    Returns
    -------
    list
        boolean list of needed parameters for the baseline
    """

    # If baseline_fx_name is not a string, extract the name from the function object
    if not isinstance(baseline_fx_name, str):
        baseline_fx_name = baseline_fx_name.__name__

    if 'constant' in baseline_fx_name:

        return [False, False]

    elif 'linear' in baseline_fx_name:

        return [True, False]

    elif 'quadratic' in baseline_fx_name:

        return [True, True]

    #elif baseline_fx_name == 'exponential':
    else:

        return [True, True]


def fit_line_robust(x,y):

    """
    Fit a line to the data using robust fitting

    Parameters
    ----------
    x : array-like
        x data
    y : array-like
        y data

    Returns
    -------
    m : float
        Slope of the fitted line
    b : float
        Intercept of the fitted line
    """

    def linear_model(x,params):
        m,b = params
        return m * x + b

    p0 = np.polyfit(x, y, 1)

    # Perform robust fitting
    res_robust = least_squares(
        lambda params: linear_model(x, params) - y,
        p0,
        loss='soft_l1',
        f_scale=0.1
    )

    m, b = res_robust.x

    return m, b

def fit_quadratic_robust(x,y):

    """
    Fit a quadratic equation to the data using robust fitting

    Parameters
    ----------
    x : array-like
        x data
    y : array-like
        y data

    Returns
    -------
    a : float
        Quadratic coefficient of the fitted polynomial
    b : float
        Linear coefficient of the fitted polynomial
    c : float
        Constant coefficient of the fitted polynomial
    """

    def model(x,params):
        a,b,c = params
        return a*np.square(x) + b*x + c

    p0 = np.polyfit(x, y, 2)

    # Perform robust fitting
    res_robust = least_squares(
        lambda params: model(x, params) - y,
        p0,
        loss='soft_l1',
        f_scale=0.1
    )

    a,b,c = res_robust.x

    return a,b,c

def fit_exponential_robust(x,y):

    """
    Fit an exponential function to the data using robust fitting.

    Notes
    -----
    Temperatures should be shifted to the reference (Tref) before calling this function.

    Parameters
    ----------
    x : array-like
        x data
    y : array-like
        y data

    Returns
    -------
    a : float
        Baseline
    c : float
        Pre-exponential factor
    alpha : float
        Exponential factor
    """

    def model(x,a,c,alpha):

        return a + c * np.exp(-alpha * x)

    # Initial parameter estimation by grid search

    rss = np.inf

    alpha_seq = np.logspace(-8, -1, 24)

    p0 = np.array( [np.min(y), np.min(y)/2])
    best_alpha = alpha_seq[0]

    low_bounds = [0, 0]

    high_bounds = [1e6, 1e6]

    for alpha in alpha_seq:

        def fit_fx(x,a,c):

            return a + c * np.exp(-alpha * x)

        params, cov = curve_fit(
            fit_fx,
            x,
            y,
            p0=p0,
            bounds=(low_bounds, high_bounds))

        pred =  fit_fx(x, *params)

        rss_curr = get_rss(y, pred)

        if rss_curr < rss:

            p0 = params
            rss = rss_curr
            best_alpha = alpha

    p0 = p0.tolist() + [best_alpha]

    low_bounds.append(0)
    high_bounds.append(1e6)

    # Perform robust fitting
    res_robust = least_squares(
        lambda params: model(x, *params) - y,
        p0,
        loss='soft_l1',
        f_scale=0.1,
        bounds=(low_bounds, high_bounds),
    )

    a,c,alpha = res_robust.x

    return a,c,alpha

def fit_thermal_unfolding(
    list_of_temperatures, 
    list_of_signals,
    initial_parameters,
    low_bounds, 
    high_bounds,
    signal_fx,
    baseline_native_fx,
    baseline_unfolded_fx,
    Cp,
    list_of_oligomer_conc=None):

    """
    Fit the thermal unfolding profile of many curves at the same time.

    This performs global fitting of shared thermodynamic parameters with per-curve baselines.

    Parameters
    ----------
    list_of_temperatures : list of array-like
        List of temperature arrays for each dataset
    list_of_signals : list of array-like
        List of signal arrays for each dataset
    initial_parameters : array-like
        Initial guess for the parameters
    low_bounds : array-like
        Lower bounds for the parameters
    high_bounds : array-like
        Upper bounds for the parameters
    signal_fx : callable
        Function to calculate the signal based on the parameters

    baseline_native_fx : callable
        function to calculate the native state baseline

    baseline_unfolded_fx : callable
        function to calculate the unfolded state baseline

    Cp : float
        Heat capacity change (passed to `signal_fx`)
    list_of_oligomer_conc : list, optional
        List of oligomer concentrations for each dataset (if applicable)

    Returns
    -------
    global_fit_params : numpy.ndarray
        Fitted global parameters
    cov : numpy.ndarray
        Covariance matrix of the fitted parameters
    predicted_lst : list of numpy.ndarray
        Predicted signals for each dataset based on the fitted parameters
    """

    all_signal = np.concatenate(list_of_signals, axis=0)

    baseline_native_params = baseline_fx_name_to_req_params(baseline_native_fx)
    baseline_unfolded_params = baseline_fx_name_to_req_params(baseline_unfolded_fx)

    list_of_temperatures = [temperature_to_kelvin(T) for T in list_of_temperatures]

    # Convert the Tm to kelvin
    initial_parameters[0] = temperature_to_kelvin(initial_parameters[0])
    low_bounds[0] = temperature_to_kelvin(low_bounds[0])
    high_bounds[0] = temperature_to_kelvin(high_bounds[0])

    def thermal_unfolding(dummyVariable, *args):

        """
        Calculate the thermal unfolding profile of many curves at the same time

        Requires:

            - The 'listOfTemperatures' containing each of them a single dataset

        The other arguments have to be in the following order:

            - Global melting temperature
            - Global enthalpy of unfolding
            - Single intercepts, folded
            - Single intercepts, unfolded
            - Single slopes or pre-exp terms, folded
            - Single slopes  or pre-exp terms, unfolded
            - Single quadratic or exponential coefficients, folded
            - Single quadratic or exponential coefficients, unfolded

        Returns:

            The melting curves based on the parameters Temperature of melting, enthalpy of unfolding,
                slopes and intercept of the folded and unfolded states

        """

        n_datasets = len(list_of_temperatures)
        Tm, dh     = args[:2]  # Temperature of melting, Enthalpy of unfolding

        intercepts_folded   = args[2:(2 + n_datasets)]
        intercepts_unfolded = args[(2 + n_datasets):(2 + n_datasets * 2)]

        id_param_init = (2 + n_datasets * 2)
        n_params      = n_datasets

        if baseline_native_params[0]:

            p2_Ns = args[id_param_init:(id_param_init+n_params)]
            id_param_init += n_params

        else:

            p2_Ns = np.zeros(n_params)

        if baseline_unfolded_params[0]:

            p2_Us = args[id_param_init:(id_param_init+n_params)]
            id_param_init += n_params

        else:

            p2_Us = np.zeros(n_params)

        if baseline_native_params[1]:

            p3_Ns = args[id_param_init:(id_param_init+n_params)]
            id_param_init += n_params

        else:

            p3_Ns = np.zeros(n_params)

        if baseline_unfolded_params[1]:

            p3_Us = args[id_param_init:(id_param_init+n_params)]
            id_param_init += n_params

        else:

            p3_Us = np.zeros(n_params)

        signal = []

        for i, T in enumerate(list_of_temperatures):

            p1_N = intercepts_folded[i]
            p1_U = intercepts_unfolded[i]

            p2_N = p2_Ns[i]
            p2_U = p2_Us[i]

            p3_N = p3_Ns[i]
            p3_U = p3_Us[i]

            y = signal_fx(
                T, Tm, dh,
                p1_N, p2_N, p3_N,
                p1_U, p2_U, p3_U,
                baseline_native_fx,
                baseline_unfolded_fx,
                Cp
            )
            signal.append(y)

        return np.concatenate(signal, axis=0)

    global_fit_params, cov = curve_fit(
        thermal_unfolding, 1, all_signal,
        p0=initial_parameters, bounds=(low_bounds, high_bounds)
        )

    predicted = thermal_unfolding(1,*global_fit_params)

    # Convert predict to list of lists
    predicted_lst = []

    init = 0
    for T in list_of_temperatures:
        n = len(T)
        predicted_lst.append(predicted[init:init+n])
        init += n

    # Convert the Tm to Celsius
    global_fit_params[0] = temperature_to_celsius(global_fit_params[0])

    return global_fit_params, cov, predicted_lst


def fit_tc_unfolding_single_slopes(
    list_of_temperatures,
    list_of_signals,
    denaturant_concentrations,
    initial_parameters,
    low_bounds,
    high_bounds,
    signal_fx,
    baseline_native_fx,
    baseline_unfolded_fx,
    list_of_oligomer_conc=None,
    fit_m1=False,
    cp_value=None,
    tm_value=None,
    dh_value=None,
    method='least_sq'
):
    """
    Vectorized and optimized version of global thermal unfolding fitting.

        Parameters
    ----------
    list_of_temperatures : list of array-like
        Temperature arrays for each dataset
    list_of_signals : list of array-like
        Signal arrays for each dataset
    denaturant_concentrations : list
        Denaturant concentrations (one per dataset)
    initial_parameters : array-like
        Initial guess for parameters
    low_bounds : array-like
        Lower bounds for parameters
    high_bounds : array-like
        Upper bounds for parameters
    signal_fx : callable
        Signal model function
    baseline_native_fx : callable
        function to calculate the native state baseline
    baseline_unfolded_fx : callable
        function to calculate the unfolded state baseline
    list_of_oligomer_conc : list, optional
        Oligomer concentrations per dataset
    fit_m1 : bool, optional
        Whether to fit temperature dependence of m-value
    cp_value, tm_value, dh_value : float or None, optional
        Optional fixed thermodynamic parameters
    method : str, optional
        Optimization method ('least_sq' or 'curve_fit')

    Returns
    -------
    global_fit_params : numpy.ndarray
         Fitted global parameters
    cov : numpy.ndarray
        Covariance matrix
    predicted_lst : list of numpy.ndarray
        Predicted signals per dataset
    """

    # ------------------------------------------------------------
    # Precompute dataset structure
    # ------------------------------------------------------------
    n_datasets = len(list_of_temperatures)
    lengths = np.array([len(T) for T in list_of_temperatures])

    list_of_temperatures = [temperature_to_kelvin(T) for T in list_of_temperatures]

    T_all = np.concatenate(list_of_temperatures)
    y_all = np.concatenate(list_of_signals)

    d_all = np.repeat(denaturant_concentrations, lengths)

    c_all = 0 if list_of_oligomer_conc is None else np.repeat(denaturant_concentrations, lengths)

    # ------------------------------------------------------------
    # Baseline parameter requirements (resolved ONCE)
    # ------------------------------------------------------------
    use_p2N, use_p3N = baseline_fx_name_to_req_params(baseline_native_fx)
    use_p2U, use_p3U = baseline_fx_name_to_req_params(baseline_unfolded_fx)

    # Convert the Tm to kelvin
    if tm_value is None:
        initial_parameters[0] = temperature_to_kelvin(initial_parameters[0])
        low_bounds[0] = temperature_to_kelvin(low_bounds[0])
        high_bounds[0] = temperature_to_kelvin(high_bounds[0])
    else:
        tm_value = temperature_to_kelvin(tm_value)

    # ------------------------------------------------------------
    # Vectorized unfolding model
    # ------------------------------------------------------------
    def unfolding(_, *params):

        i = 0

        # ---- Global thermodynamics ----
        if tm_value is None:
            Tm = params[i]
            i += 1
        else:
            Tm = tm_value

        if dh_value is None:
            DHm = params[i]
            i += 1
        else:
            DHm = dh_value

        if cp_value is None:
            Cp0 = params[i]
            i += 1
        else:
            Cp0 = cp_value

        m0 = params[i]
        i += 1

        if fit_m1:
            m1 = params[i]
            i += 1
        else:
            m1 = 0.0

        # ---- Dataset-specific parameters ----
        p1N = np.repeat(params[i:i + n_datasets], lengths)
        i += n_datasets

        p1U = np.repeat(params[i:i + n_datasets], lengths)
        i += n_datasets

        if use_p2N:
            p2N = np.repeat(params[i:i + n_datasets], lengths)
            i += n_datasets
        else:
            p2N = 0.0

        if use_p2U:
            p2U = np.repeat(params[i:i + n_datasets], lengths)
            i += n_datasets
        else:
            p2U = 0.0

        if use_p3N:
            p3N = np.repeat(params[i:i + n_datasets], lengths)
            i += n_datasets
        else:
            p3N = 0.0

        if use_p3U:
            p3U = np.repeat(params[i:i + n_datasets], lengths)
            i += n_datasets
        else:
            p3U = 0.0

        # ---- Single vectorized signal evaluation ----
        return signal_fx(
            T_all, d_all,
            DHm, Tm, Cp0, m0, m1,
            0, p1N, p2N, p3N,
            0, p1U, p2U, p3U,
            baseline_native_fx,
            baseline_unfolded_fx,
            c_all
        )

    if method == 'least_sq':

        def residuals(params):
            return unfolding(None, *params) - y_all

        res = least_squares(
            residuals,
            x0=initial_parameters,
            bounds=(low_bounds, high_bounds),
            method="trf",  # trust-region reflective (supports bounds)
            xtol=1e-8,
            ftol=1e-8,
            gtol=1e-8
        )

        global_fit_params = res.x

        # Jacobian at solution
        J = res.jac

        # Residual variance
        dof = len(y_all) - len(res.x)
        residual_variance = np.sum(res.fun ** 2) / dof

        # Robust covariance matrix
        cov = pinv(J.T @ J) * residual_variance

    else:

        # ------------------------------------------------------------
        # Fit
        # ------------------------------------------------------------
        global_fit_params, cov = curve_fit(
            unfolding,
            xdata=1.0,            # dummy variable
            ydata=y_all,
            p0=initial_parameters,
            bounds=(low_bounds, high_bounds)
        )

    # ------------------------------------------------------------
    # Predict & split per dataset
    # ------------------------------------------------------------
    predicted_all = unfolding(1.0, *global_fit_params)

    predicted_lst = []
    start = 0
    for n in lengths:
        predicted_lst.append(predicted_all[start:start + n])
        start += n

    # Convert the Tm back to Celsius
    if tm_value is None:
        global_fit_params[0] = temperature_to_celsius(global_fit_params[0])

    return global_fit_params, cov, predicted_lst

def fit_oligomer_unfolding_single_slopes(
        list_of_temperatures,
        list_of_signals,
        oligomer_concentrations,
        initial_parameters,
        low_bounds,
        high_bounds,
        signal_fx,
        baseline_native_fx,
        baseline_unfolded_fx,
        cp_value=None,
        tm_value=None,
        dh_value=None,
):
    """
    Vectorized and optimized version of global thermal unfolding fitting. of oligomers

    Parameters
    ----------
    list_of_temperatures : list of array-like
        Temperature arrays for each dataset
    list_of_signals : list of array-like
        Signal arrays for each dataset
    oligomer_concentrations : list
        sample concentrations of the oligomeric complex (one per dataset)
    initial_parameters : array-like
        Initial guess for parameters
    low_bounds : array-like
        Lower bounds for parameters
    high_bounds : array-like
        Upper bounds for parameters
    signal_fx : callable
        Signal model function
    baseline_native_fx : callable
        function to calculate the native state baseline
    baseline_unfolded_fx : callable
        function to calculate the unfolded state baseline
    cp_value, tm_value, dh_value : float or None, optional
        Optional fixed thermodynamic parameters

    Returns
    -------
    global_fit_params : numpy.ndarray
         Fitted global parameters
    cov : numpy.ndarray
        Covariance matrix
    predicted_lst : list of numpy.ndarray
        Predicted signals per dataset
    """

    # ------------------------------------------------------------
    # Precompute dataset structure
    # ------------------------------------------------------------
    n_datasets = len(list_of_temperatures)
    lengths = np.array([len(T) for T in list_of_temperatures])

    list_of_temperatures = [temperature_to_kelvin(T) for T in list_of_temperatures]

    T_all = np.concatenate(list_of_temperatures)
    y_all = np.concatenate(list_of_signals)

    C_all = np.repeat(oligomer_concentrations, lengths)

    # ------------------------------------------------------------
    # Baseline parameter requirements (resolved ONCE)
    # ------------------------------------------------------------
    use_p2N, use_p3N = baseline_fx_name_to_req_params(baseline_native_fx)
    use_p2U, use_p3U = baseline_fx_name_to_req_params(baseline_unfolded_fx)

    # Convert the Tm to kelvin
    if tm_value is None:
        initial_parameters[0] = temperature_to_kelvin(initial_parameters[0])
        low_bounds[0] = temperature_to_kelvin(low_bounds[0])
        high_bounds[0] = temperature_to_kelvin(high_bounds[0])
    else:
        tm_value = temperature_to_kelvin(tm_value)

    def unfolding(_, *params):

        """
        Calculate the thermal unfolding profile of many curves at the same time

        Requires:

            - The 'T_all' containing the temperatures as a single dataset
            - The 'C_all' containing the concentrations as a single dataset

        The other arguments have to be in the following order:

            - Global melting temperature
            - Global enthalpy of unfolding
            - Global Cp0
            - Single intercepts, folded
            - Single intercepts, unfolded
            - Single slopes or pre-exp terms, folded
            - Single slopes or pre-exp terms, unfolded
            - Single quadratic or exponential coefficients, folded
            - Single quadratic or exponential coefficients, unfolded

        Returns:

            The melting curves based on the parameters Temperature of melting, enthalpy of unfolding,
                slopes and intercept of the folded and unfolded states

        """
        i = 0

        # ---- Global thermodynamics ----
        if tm_value is None:
            Tm = params[i]
            i += 1
        else:
            Tm = tm_value

        if dh_value is None:
            DHm = params[i]
            i += 1
        else:
            DHm = dh_value

        if cp_value is None:
            Cp0 = params[i]
            i += 1
        else:
            Cp0 = cp_value


        # ---- Dataset-specific parameters ----
        p1N = np.repeat(params[i:i + n_datasets], lengths)
        i += n_datasets

        p1U = np.repeat(params[i:i + n_datasets], lengths)
        i += n_datasets

        if use_p2N:
            p2N = np.repeat(params[i:i + n_datasets], lengths)
            i += n_datasets
        else:
            p2N = 0.0

        if use_p2U:
            p2U = np.repeat(params[i:i + n_datasets], lengths)
            i += n_datasets
        else:
            p2U = 0.0

        if use_p3N:
            p3N = np.repeat(params[i:i + n_datasets], lengths)
            i += n_datasets
        else:
            p3N = 0.0

        if use_p3U:
            p3U = np.repeat(params[i:i + n_datasets], lengths)
            i += n_datasets
        else:
            p3U = 0.0

        # ---- Single vectorized signal evaluation ----
        return signal_fx(
                T_all,C_all, Tm, DHm,
                0, p1N, p2N, p3N,
                0, p1U, p2U, p3U,
                baseline_native_fx,
                baseline_unfolded_fx,
                Cp0,
            )


    global_fit_params, cov = curve_fit(
        unfolding, 1.0, y_all,
        p0=initial_parameters, bounds=(low_bounds, high_bounds)
    )

    predicted_all = unfolding(1.0, *global_fit_params)

    predicted_lst = []
    start = 0
    for n in lengths:
        predicted_lst.append(predicted_all[start:start + n])
        start += n

    # Convert the Tm back to Celsius
    if tm_value is None:
        global_fit_params[0] = temperature_to_celsius(global_fit_params[0])

    return global_fit_params, cov, predicted_lst


def fit_tc_unfolding_shared_slopes_many_signals(
    list_of_temperatures,
    list_of_signals,
    signal_ids,
    denaturant_concentrations,
    initial_parameters,
    low_bounds,
    high_bounds,
    signal_fx,
    baseline_native_fx,
    baseline_unfolded_fx,
    list_of_oligomer_conc=None,
    fit_m1=False,
    cp_value=None,
    tm_value=None,
    dh_value=None
):
    """
    Vectorized fitting of thermochemical unfolding curves for multiple signal types
    sharing thermodynamic parameters and slopes, using least_squares.

    Parameters
    ----------
    list_of_temperatures : list of array-like
        Temperature arrays for each dataset
    list_of_signals : list of array-like
        Signal arrays for each dataset
    signal_ids : list of int
        Signal-type id for each dataset (0..n_signals-1)
    denaturant_concentrations : list
        Denaturant concentrations for each dataset (flattened across signals)
    initial_parameters : array-like
        Initial guess for the parameters
    low_bounds : array-like
        Lower bounds for the parameters
    high_bounds : array-like
        Upper bounds for the parameters
    signal_fx : callable
        Signal model function
     baseline_native_fx : callable
        function to calculate the baseline for the native state
    baseline_unfolded_fx : callable
        function to calculate the baseline for the unfolded state
    list_of_oligomer_conc : list, optional
        Oligomer concentrations per dataset
    fit_m1 : bool, optional
        Whether to fit temperature dependence of m-value
    cp_value, tm_value, dh_value : float or None, optional
        Optional fixed thermodynamic parameters

    Returns
    -------
    global_fit_params : numpy.ndarray
         Fitted global parameters
    cov : numpy.ndarray
        Covariance matrix
    predicted_lst : list of numpy.ndarray
        Predicted signals per dataset
    """

    # Flatten all signals
    all_signal = np.concatenate(list_of_signals, axis=0)
    n_signals = np.max(signal_ids) + 1
    n_datasets = len(list_of_temperatures)

    list_of_temperatures = [temperature_to_kelvin(T) for T in list_of_temperatures]

    baseline_native_params = baseline_fx_name_to_req_params(baseline_native_fx)
    baseline_unfolded_params = baseline_fx_name_to_req_params(baseline_unfolded_fx)

    # Precompute indices for slicing the flattened concatenated arrays
    dataset_starts = np.cumsum([0] + [len(T) for T in list_of_temperatures][:-1])
    dataset_ends = np.cumsum([len(T) for T in list_of_temperatures])

    # Convert the Tm to kelvin
    if tm_value is None:
        initial_parameters[0] = temperature_to_kelvin(initial_parameters[0])
        low_bounds[0] = temperature_to_kelvin(low_bounds[0])
        high_bounds[0] = temperature_to_kelvin(high_bounds[0])
    else:
        tm_value = temperature_to_kelvin(tm_value)

    # Vectorized residuals function for least_squares
    def residuals(params):
        id_param = 0

        Tm = params[id_param] if tm_value is None else tm_value
        if tm_value is None:
            id_param += 1

        DHm = params[id_param] if dh_value is None else dh_value
        if dh_value is None:
            id_param += 1

        Cp0 = params[id_param] if cp_value is None else cp_value
        if cp_value is None:
            id_param += 1

        m0 = params[id_param]
        id_param += 1

        m1 = params[id_param] if fit_m1 else 0
        if fit_m1:
            id_param += 1

        intercepts_folded = params[id_param:id_param + n_datasets]
        intercepts_unfolded = params[id_param + n_datasets:id_param + n_datasets * 2]
        id_param += n_datasets * 2

        # Shared slopes / coefficients per signal type
        p2_n_s = params[id_param:id_param + n_signals] if baseline_native_params[0] else np.zeros(n_signals)
        id_param += n_signals if baseline_native_params[0] else 0

        p2_u_s = params[id_param:id_param + n_signals] if baseline_unfolded_params[0] else np.zeros(n_signals)
        id_param += n_signals if baseline_unfolded_params[0] else 0

        p3_n_s = params[id_param:id_param + n_signals] if baseline_native_params[1] else np.zeros(n_signals)
        id_param += n_signals if baseline_native_params[1] else 0

        p3_u_s = params[id_param:id_param + n_signals] if baseline_unfolded_params[1] else np.zeros(n_signals)
        id_param += n_signals if baseline_unfolded_params[1] else 0

        # Vectorized evaluation for all datasets
        predicted_all = np.zeros_like(all_signal)
        for i, T in enumerate(list_of_temperatures):
            start, end = dataset_starts[i], dataset_ends[i]
            d = denaturant_concentrations[i]
            c = 0 if list_of_oligomer_conc is None else list_of_oligomer_conc[i]
            sig_id = signal_ids[i]

            predicted_all[start:end] = signal_fx(
                T, d, DHm, Tm, Cp0, m0, m1,
                0, intercepts_folded[i], p2_n_s[sig_id], p3_n_s[sig_id],
                0, intercepts_unfolded[i], p2_u_s[sig_id], p3_u_s[sig_id],
                baseline_native_fx,
                baseline_unfolded_fx,
                c
            )

        return predicted_all - all_signal

    # Run least_squares fit
    res = least_squares(
        residuals,
        x0=initial_parameters,
        bounds=(low_bounds, high_bounds),
        method="trf",
        max_nfev=2000
    )

    global_fit_params = res.x

    # Compute robust covariance using pseudo-inverse
    J = res.jac
    dof = len(all_signal) - len(global_fit_params)
    residual_variance = np.sum(res.fun**2) / dof
    cov = np.linalg.pinv(J.T @ J) * residual_variance

    # Convert predicted signal into list of arrays per dataset
    predicted = res.fun + all_signal
    predicted_lst = [predicted[start:end] for start, end in zip(dataset_starts, dataset_ends)]

    # Convert the Tm back to Celsius
    if tm_value is None:
        global_fit_params[0] = temperature_to_celsius(global_fit_params[0])

    return global_fit_params, cov, predicted_lst

def fit_oligomer_unfolding_shared_slopes_many_signals(
    list_of_temperatures,
    list_of_signals,
    signal_ids,
    oligomer_concentrations,
    initial_parameters,
    low_bounds,
    high_bounds,
    signal_fx,
    baseline_native_fx,
    baseline_unfolded_fx,
    cp_value=None,
    tm_value=None,
    dh_value=None
):
    """
    Vectorized fitting of oligomer thermal unfolding curves for multiple signal types
    sharing thermodynamic parameters and slopes, using least_squares.

    Parameters
    ----------
    list_of_temperatures : list of array-like
        Temperature arrays for each dataset.
    list_of_signals : list of array-like
        Signal arrays for each dataset.
    signal_ids : list of int
        Signal-type id for each dataset (0..n_signals-1)
    oligomer_concentrations : list
        sample concentrations of the oligomeric complex for each dataset (flattened across signals)
    initial_parameters : array-like
        Initial guess for the parameters
    low_bounds : array-like
        Lower bounds for the parameters
    high_bounds : array-like
        Upper bounds for the parameters
    signal_fx : callable
        Signal model function
     baseline_native_fx : callable
        function to calculate the baseline for the native state
    baseline_unfolded_fx : callable
        function to calculate the baseline for the unfolded state
    cp_value, tm_value, dh_value : float or None, optional
        Optional fixed thermodynamic parameters

    Returns
    -------
    global_fit_params : numpy.ndarray
         Fitted global parameters
    cov : numpy.ndarray
        Covariance matrix
    predicted_lst : list of numpy.ndarray
        Predicted signals per dataset

    """

    # Flatten all signals
    all_signal = np.concatenate(list_of_signals, axis=0)
    n_signals = np.max(signal_ids) + 1
    n_datasets = len(list_of_temperatures)

    list_of_temperatures = [temperature_to_kelvin(T) for T in list_of_temperatures]

    baseline_native_params = baseline_fx_name_to_req_params(baseline_native_fx)
    baseline_unfolded_params = baseline_fx_name_to_req_params(baseline_unfolded_fx)

    # Precompute indices for slicing the flattened concatenated arrays
    dataset_starts = np.cumsum([0] + [len(T) for T in list_of_temperatures][:-1])
    dataset_ends = np.cumsum([len(T) for T in list_of_temperatures])

    # Convert the Tm to kelvin
    if tm_value is None:
        initial_parameters[0] = temperature_to_kelvin(initial_parameters[0])
        low_bounds[0] = temperature_to_kelvin(low_bounds[0])
        high_bounds[0] = temperature_to_kelvin(high_bounds[0])
    else:
        tm_value = temperature_to_kelvin(tm_value)

    # Vectorized residuals function for least_squares
    def residuals(params):
        """
        Calculate the thermal unfolding profile of many curves at the same time

        Requires:

            - The 'listOfTemperatures' containing each of them a single dataset

        The other arguments have to be in the following order:

            - Global melting temperature
            - Global enthalpy of unfolding
            - Global Cp0
            - Shared intercept, folded
            - Shared intercept, unfolded
            - Single slopes or pre-exp terms, folded
            - Single slopes or pre-exp terms, unfolded
            - Single quadratic or exponential coefficients, folded
            - Single quadratic or exponential coefficients, unfolded

        Returns:

            The melting curves based on the parameters Temperature of melting, enthalpy of unfolding,
                slopes and intercept of the folded and unfolded states

        """
        id_param = 0

        Tm = params[id_param] if tm_value is None else tm_value
        if tm_value is None:
            id_param += 1

        DHm = params[id_param] if dh_value is None else dh_value
        if dh_value is None:
            id_param += 1

        Cp0 = params[id_param] if cp_value is None else cp_value
        if cp_value is None:
            id_param += 1


        intercepts_folded = params[id_param:id_param + n_datasets]
        intercepts_unfolded = params[id_param + n_datasets:id_param + n_datasets * 2]
        id_param += n_datasets * 2

        # Shared slopes / coefficients per signal type
        p2_n_s = params[id_param:id_param + n_signals] if baseline_native_params[0] else np.zeros(n_signals)
        id_param += n_signals if baseline_native_params[0] else 0

        p2_u_s = params[id_param:id_param + n_signals] if baseline_unfolded_params[0] else np.zeros(n_signals)
        id_param += n_signals if baseline_unfolded_params[0] else 0

        p3_n_s = params[id_param:id_param + n_signals] if baseline_native_params[1] else np.zeros(n_signals)
        id_param += n_signals if baseline_native_params[1] else 0

        p3_u_s = params[id_param:id_param + n_signals] if baseline_unfolded_params[1] else np.zeros(n_signals)
        id_param += n_signals if baseline_unfolded_params[1] else 0

        # Vectorized evaluation for all datasets
        predicted_all = np.zeros_like(all_signal)
        for i, T in enumerate(list_of_temperatures):
            start, end = dataset_starts[i], dataset_ends[i]
            c = oligomer_concentrations[i]
            sig_id = signal_ids[i]

            predicted_all[start:end] = signal_fx(
                T, c, Tm, DHm,
                0, intercepts_folded[i], p2_n_s[sig_id], p3_n_s[sig_id],
                0, intercepts_unfolded[i], p2_u_s[sig_id], p3_u_s[sig_id],
                baseline_native_fx,
                baseline_unfolded_fx,
                Cp0
            )


        return predicted_all - all_signal

    # Run least_squares fit
    res = least_squares(
        residuals,
        x0=initial_parameters,
        bounds=(low_bounds, high_bounds),
        method="trf",
        max_nfev=2000
    )

    global_fit_params = res.x

    # Compute robust covariance using pseudo-inverse
    J = res.jac
    dof = len(all_signal) - len(global_fit_params)
    residual_variance = np.sum(res.fun**2) / dof
    cov = np.linalg.pinv(J.T @ J) * residual_variance

    # Convert predicted signal into list of arrays per dataset
    predicted = res.fun + all_signal
    predicted_lst = [predicted[start:end] for start, end in zip(dataset_starts, dataset_ends)]

    # Convert the Tm back to Celsius
    if tm_value is None:
        global_fit_params[0] = temperature_to_celsius(global_fit_params[0])

    return global_fit_params, cov, predicted_lst

def fit_tc_unfolding_many_signals(
        list_of_temperatures,
        list_of_signals,
        signal_ids,
        denaturant_concentrations,
        initial_parameters,
        low_bounds, high_bounds,
        signal_fx,
        baseline_native_fx,
        baseline_unfolded_fx,
        oligomer_concentrations=None,
        fit_m1=False,
        model_scale_factor=False,
        scale_factor_exclude_ids=[],
        cp_value=None,
        fit_native_den_slope=True,
        fit_unfolded_den_slope=True):
    """
    Fit thermochemical unfolding curves for many signals (optimized variant).

    Parameters
    ----------
    list_of_temperatures : list of array-like
        Temperature arrays for each dataset.
    list_of_signals : list of array-like
        Signal arrays for each dataset.
    signal_ids : list of int
        Signal-type id for each dataset (0..n_signals-1)
    denaturant_concentrations : list
        Denaturant concentrations for each dataset (flattened across signals)
    initial_parameters : array-like
        Initial guess for the parameters
    low_bounds : array-like
        Lower bounds for the parameters
    high_bounds : array-like
        Upper bounds for the parameters
    signal_fx : callable
        Signal model function
    baseline_native_fx : callable
        function to calculate the native state baseline
    baseline_unfolded_fx : callable
        function to calculate the unfolded state baseline
    oligomer_concentrations : list, optional
        Oligomer concentrations per dataset (used by oligomeric models)
    fit_m1 : bool, optional
        Whether to include and fit temperature dependence of the m-value (m1)
    model_scale_factor : bool, optional
        If True, include a per-denaturant concentration scale factor to account for intensity differences
    scale_factor_exclude_ids : list, optional
        IDs of scale factors to exclude / fix to 1
    cp_value : float or None, optional
        If provided, Cp is fixed to this value and not fitted
    fit_native_den_slope, fit_unfolded_den_slope : bool, optional
        Whether to fit denaturant dependence of baselines.

    Returns
    -------
    global_fit_params : numpy.ndarray
         Fitted global parameters
    cov : numpy.ndarray
        Covariance matrix
    predicted_lst : list of numpy.ndarray
        Predicted signals per dataset
    """

    all_signal = np.concatenate(list_of_signals, axis=0)

    n_signals = np.max(signal_ids) + 1

    nr_den = int(len(denaturant_concentrations) / n_signals)

    if len(scale_factor_exclude_ids) > 0 and model_scale_factor:
        # Sort them in ascending order to avoid issues when inserting
        scale_factor_exclude_ids = sorted(scale_factor_exclude_ids)

    baseline_native_params = [fit_native_den_slope] + baseline_fx_name_to_req_params(baseline_native_fx)
    baseline_unfolded_params = [fit_unfolded_den_slope] + baseline_fx_name_to_req_params(baseline_unfolded_fx)

    initial_parameters[0] = temperature_to_kelvin(initial_parameters[0])
    low_bounds[0] = temperature_to_kelvin(low_bounds[0])
    high_bounds[0] = temperature_to_kelvin(high_bounds[0])

    list_of_temperatures = [temperature_to_kelvin(T) for T in list_of_temperatures]

    def unfolding(dummyVariable, *args):

        """
        The parameters order is as follows:

            Tm, Dh, Cp0 and m-value

            Intercept folded
            Intercept unfolded

            Temperature slope or term pre-exponential factor folded
            Temperature slope term or pre-exponential factor unfolded

            Denaturant slope term folded
            Denaturant slope term unfolded

            Quadratic coefficient or exponential coefficient folded
            Quadratic coefficient or exponential coefficient unfolded

        """

        if cp_value is not None:

            Cp0 = cp_value
            Tm, DHm, m0 = args[:3]  # Enthalpy of unfolding, Temperature of melting, m0, m1
            id_param_init = 3

        else:

            Tm, DHm, Cp0, m0 = args[:4]  # Enthalpy of unfolding, Temperature of melting, Cp0, m0, m1
            id_param_init = 4

        if fit_m1:
            m1 = args[id_param_init]
            id_param_init += 1
        else:
            m1 = 0

        # Intercept parameters
        p2_Ns = args[id_param_init:id_param_init + n_signals]
        p2_Us = args[id_param_init + n_signals:id_param_init + 2 * n_signals]

        id_param_init = id_param_init + 2 * n_signals

        # Temperature slope or pre-exponential parameters
        if baseline_native_params[1]:

            p3_Ns = args[id_param_init:id_param_init + n_signals]
            id_param_init += n_signals

        else:
            p3_Ns = [0] * n_signals

        if baseline_unfolded_params[1]:

            p3_Us = args[id_param_init:id_param_init + n_signals]
            id_param_init += n_signals

        else:
            p3_Us = [0] * n_signals

        # Denaturant slope parameters
        if baseline_native_params[0]:

            p1_Ns = args[id_param_init:id_param_init + n_signals]
            id_param_init += n_signals

        else:

            p1_Ns = [0] * n_signals

        if baseline_unfolded_params[0]:

            p1_Us = args[id_param_init:id_param_init + n_signals]
            id_param_init += n_signals

        else:

            p1_Us = [0] * n_signals

        # Temperature-dependent quadratic or exponential coefficients
        if baseline_native_params[2]:

            p4_Ns = args[id_param_init:id_param_init + n_signals]
            id_param_init += n_signals

        else:
            p4_Ns = [0] * n_signals

        if baseline_unfolded_params[2]:

            p4_Us = args[id_param_init:id_param_init + n_signals]
            id_param_init += n_signals

        else:
            p4_Us = [0] * n_signals

        if model_scale_factor:
            # One per denaturant concentration
            factors = args[id_param_init:id_param_init + (nr_den - len(scale_factor_exclude_ids))]

            for id_ex in scale_factor_exclude_ids:
                factors = np.insert(factors, id_ex, 1)

            # Repeat the list so have the same length as list_of_temperatures, equal to denaturant concentration * number of signals
            factors = np.tile(factors, n_signals)

            id_param_init += nr_den

        signal = []

        for i, T in enumerate(list_of_temperatures):

            p1_N = p1_Ns[signal_ids[i]]
            p1_U = p1_Us[signal_ids[i]]
            p2_N = p2_Ns[signal_ids[i]]
            p2_U = p2_Us[signal_ids[i]]
            p3_N = p3_Ns[signal_ids[i]]
            p3_U = p3_Us[signal_ids[i]]
            p4_N = p4_Ns[signal_ids[i]]
            p4_U = p4_Us[signal_ids[i]]

            d = denaturant_concentrations[i]

            c = 0 if oligomer_concentrations is None else oligomer_concentrations[i]

            y = signal_fx(
                T, d, DHm, Tm, Cp0, m0, m1,
                p1_N, p2_N, p3_N, p4_N,
                p1_U, p2_U, p3_U, p4_U,
                baseline_native_fx,
                baseline_unfolded_fx,
                c
            )

            scale_factor = 1 if not model_scale_factor else factors[i]

            y = y * scale_factor

            signal.append(y)

        return np.concatenate(signal, axis=0)

    global_fit_params, cov = curve_fit(
        unfolding, 1, all_signal,
        p0=initial_parameters,
        bounds=(low_bounds, high_bounds))

    predicted = unfolding(1, *global_fit_params)

    # Convert predict to list of lists
    predicted_lst = []

    init = 0
    for T in list_of_temperatures:
        n = len(T)
        predicted_lst.append(predicted[init:init + n])
        init += n

    # Convert the Tm to Celsius
    global_fit_params[0] = temperature_to_celsius(global_fit_params[0])

    return global_fit_params, cov, predicted_lst

def fit_oligomer_unfolding_many_signals(
        list_of_temperatures,
        list_of_signals,
        signal_ids,
        oligomer_concentrations,
        initial_parameters,
        low_bounds, high_bounds,
        signal_fx,
        baseline_native_fx,
        baseline_unfolded_fx,
        model_scale_factor=False,
        scale_factor_exclude_ids=[],
        cp_value=None,
        fit_native_olig_slope=True,
        fit_unfolded_olig_slope=True):
    """
    Fit thermal unfolding curves of oligomers for many signals (optimized variant).

    Parameters
    ----------
    list_of_temperatures : list of array-like
        Temperature arrays for each dataset
    list_of_signals : list of array-like
        Signal arrays for each dataset
    signal_ids : list of int
        Signal-type id for each dataset (0..n_signals-1)
    oligomer_concentrations : list
        sample concentrations of the oligomeric complex for each dataset (flattened across signals)
    initial_parameters : array-like
        Initial guess for the parameters
    low_bounds : array-like
        Lower bounds for the parameters
    high_bounds : array-like
        Upper bounds for the parameters
    signal_fx : callable
        Signal model function
    baseline_native_fx : callable
        function to calculate the native state baseline
    baseline_unfolded_fx : callable
        function to calculate the unfolded state baseline
    model_scale_factor : bool, optional
        If True, include a per-oligomeric concentration scale factor to account for intensity differences
    scale_factor_exclude_ids : list, optional
        IDs of scale factors to exclude / fix to 1
    cp_value : float or None, optional
        If provided, Cp is fixed to this value and not fitted
    fit_native_olig_slope, fit_unfolded_olig_slope : bool, optional
        Whetever to fit the dependence of the slopes of the baselines

   Returns
    -------
    global_fit_params : numpy.ndarray
         Fitted global parameters
    cov : numpy.ndarray
        Covariance matrix
    predicted_lst : list of numpy.ndarray
        Predicted signals per dataset
    """

    all_signal = np.concatenate(list_of_signals, axis=0)

    n_signals = np.max(signal_ids) + 1

    nr_olig = int(len(oligomer_concentrations) / n_signals)

    if len(scale_factor_exclude_ids) > 0 and model_scale_factor:
        # Sort them in ascending order to avoid issues when inserting
        scale_factor_exclude_ids = sorted(scale_factor_exclude_ids)

    baseline_native_params = [fit_native_olig_slope] + baseline_fx_name_to_req_params(baseline_native_fx)
    baseline_unfolded_params = [fit_unfolded_olig_slope] + baseline_fx_name_to_req_params(baseline_unfolded_fx)

    initial_parameters[0] = temperature_to_kelvin(initial_parameters[0])
    low_bounds[0] = temperature_to_kelvin(low_bounds[0])
    high_bounds[0] = temperature_to_kelvin(high_bounds[0])

    list_of_temperatures = [temperature_to_kelvin(T) for T in list_of_temperatures]

    def unfolding(dummyVariable, *args):

        """
        The parameters order is as follows:

            Tm, Dh, Cp0

            Intercept folded
            Intercept unfolded

            Temperature slope or term pre-exponential factor folded
            Temperature slope term or pre-exponential factor unfolded

            Oligomer slope term folded
            Oligomer slope term unfolded

            Quadratic coefficient or exponential coefficient folded
            Quadratic coefficient or exponential coefficient unfolded

        """

        if cp_value is not None:

            Cp0 = cp_value
            Tm, DHm = args[:2]  # Enthalpy of unfolding, Temperature of melting
            id_param_init = 2

        else:

            Tm, DHm, Cp0 = args[:3]  # Enthalpy of unfolding, Temperature of melting, Cp0
            id_param_init = 3

        # Intercept parameters
        p2_Ns = args[id_param_init:id_param_init + n_signals]
        p2_Us = args[id_param_init + n_signals:id_param_init + 2 * n_signals]

        id_param_init = id_param_init + 2 * n_signals

        # Temperature slope or pre-exponential parameters
        if baseline_native_params[1]:

            p3_Ns = args[id_param_init:id_param_init + n_signals]
            id_param_init += n_signals

        else:
            p3_Ns = [0] * n_signals

        if baseline_unfolded_params[1]:

            p3_Us = args[id_param_init:id_param_init + n_signals]
            id_param_init += n_signals

        else:
            p3_Us = [0] * n_signals

        # Oligomer slope parameters
        if baseline_native_params[0]:

            p1_Ns = args[id_param_init:id_param_init + n_signals]
            id_param_init += n_signals

        else:

            p1_Ns = [0] * n_signals

        if baseline_unfolded_params[0]:

            p1_Us = args[id_param_init:id_param_init + n_signals]
            id_param_init += n_signals

        else:

            p1_Us = [0] * n_signals

        # Temperature-dependent quadratic or exponential coefficients
        if baseline_native_params[2]:

            p4_Ns = args[id_param_init:id_param_init + n_signals]
            id_param_init += n_signals

        else:
            p4_Ns = [0] * n_signals

        if baseline_unfolded_params[2]:

            p4_Us = args[id_param_init:id_param_init + n_signals]
            id_param_init += n_signals

        else:
            p4_Us = [0] * n_signals

        if model_scale_factor:
            # One per oligomer concentration
            factors = args[id_param_init:id_param_init + (nr_olig - len(scale_factor_exclude_ids))]

            for id_ex in scale_factor_exclude_ids:
                factors = np.insert(factors, id_ex, 1)

            # Repeat the list so have the same length as list_of_temperatures, equal to denaturant concentration * number of signals
            factors = np.tile(factors, n_signals)

            id_param_init += nr_olig

        signal = []

        for i, T in enumerate(list_of_temperatures):

            p1_N = p1_Ns[signal_ids[i]]
            p1_U = p1_Us[signal_ids[i]]
            p2_N = p2_Ns[signal_ids[i]]
            p2_U = p2_Us[signal_ids[i]]
            p3_N = p3_Ns[signal_ids[i]]
            p3_U = p3_Us[signal_ids[i]]
            p4_N = p4_Ns[signal_ids[i]]
            p4_U = p4_Us[signal_ids[i]]

            c = oligomer_concentrations[i]

            y = signal_fx(
                T, c, Tm, DHm,
                p1_N, p2_N, p3_N, p4_N,
                p1_U, p2_U, p3_U, p4_U,
                baseline_native_fx,
                baseline_unfolded_fx,
                Cp0
            )


            scale_factor = 1 if not model_scale_factor else factors[i]

            y = y * scale_factor

            signal.append(y)

        return np.concatenate(signal, axis=0)

    global_fit_params, cov = curve_fit(
        unfolding, 1.0, all_signal,
        p0=initial_parameters,
        bounds=(low_bounds, high_bounds))

    predicted = unfolding(1.0, *global_fit_params)

    # Convert predict to list of lists
    predicted_lst = []

    init = 0
    for T in list_of_temperatures:
        n = len(T)
        predicted_lst.append(predicted[init:init + n])
        init += n

    # Convert the Tm to Celsius
    global_fit_params[0] = temperature_to_celsius(global_fit_params[0])

    return global_fit_params, cov, predicted_lst

def evaluate_need_to_refit(
        global_fit_params,
        high_bounds,
        low_bounds,
        p0,
        fit_m1=False,
        check_cp=True,
        check_dh=True,
        check_tm=True,
        fixed_cp=False,
        threshold=0.05,
        fit_m0=True
    ):

    """
    Check and expand parameter bounds when fitted parameters are too close to boundaries.

    Parameters
    ----------
    global_fit_params : array-like
        Fitted parameters
    high_bounds : array-like
        Upper bounds
    low_bounds : array-like
        Lower bounds
    p0 : array-like
        Initial guess for parameters
    fit_m1 : bool, optional
        Whether m1 (temperature dependence of m-value) is fitted
    check_cp, check_dh, check_tm : bool, optional
        Whether to check boundaries for Cp, DHm, and Tm respectively
    fixed_cp : bool, optional
        Whether the Cp value is fixed
    threshold : float, optional
        Threshold to compare if the fitted parameters are too close to the boundaries
    fit_m0 : bool, optional
        Whether m0 (m-value) is fitted (not in oligomeric models)

    Returns
    -------
    re_fit : bool
        True if a refit is recommended after bounds expansion
    p0 : array-like
        Updated initial parameters
    low_bounds : array-like
        Updated lower bounds
    high_bounds : array-like
        Updated upper bounds
    """

    # We need to create copies of the arrays, otherwise they will be overwritten
    global_fit_params = global_fit_params.copy()
    p0 = p0.copy()
    high_bounds = high_bounds.copy()
    low_bounds = low_bounds.copy()

    re_fit = False

    # Check the Tm boundary - upper
    tm_diff = high_bounds[0] - global_fit_params[0]

    # Expand the boundary if the Tm is too close to the boundary
    if tm_diff < 6 and check_tm:
        high_bounds[0] = global_fit_params[0] + 12
        p0[0] = global_fit_params[0] + 5
        re_fit = True

    # Check the Tm boundary - lower
    tm_diff = global_fit_params[0] - low_bounds[0]

    # Expand the boundary if the Tm is too close to the boundary
    if tm_diff < 6 and check_tm:
        low_bounds[0] = global_fit_params[0] - 12
        p0[0] = global_fit_params[0] - 5
        re_fit = True

    # Check the Dh boundary
    dh_diff = high_bounds[1] - global_fit_params[1]
    # Expand the boundary if the Dh is too close to the boundary
    if dh_diff < 20 and check_dh:
        high_bounds[1] = global_fit_params[1] + 80
        p0[1] = global_fit_params[1] + 50
        re_fit = True

    id_next = 2
    if not fixed_cp:

        # Check the Cp boundary
        cp_diff = high_bounds[2] - global_fit_params[2]
        # Expand the boundary if the Cp is too close to the boundary
        if cp_diff < 0.25 and check_cp:
            high_bounds[2] = global_fit_params[2] + 1
            p0[2] = global_fit_params[2] + 0.5
            re_fit = True
        
        id_next += 1

    if fit_m0:
        # Check the m-value boundary
        m_diff = high_bounds[id_next] - global_fit_params[id_next]
        # Expand the boundary if the m-value is too close to the boundary
        if m_diff < 0.5:
            high_bounds[id_next] = global_fit_params[id_next] + 2
            p0[id_next] = global_fit_params[id_next] + 0.5
            re_fit = True

        # Evaluate if m1 is fitted
        id_start = id_next + 1
    else:
        id_start = id_next


    if fit_m1:

        m1_diff = high_bounds[id_start] - global_fit_params[id_start]
        # Expand the boundary if the m-value is too close to the boundary
        if m1_diff < 0.1:
            high_bounds[id_start] = global_fit_params[id_start] + 1
            re_fit = True

        m1_diff = global_fit_params[id_start] - low_bounds[id_start]
        # Expand the boundary if the m-value is too close to the boundary
        if m1_diff < 0.1:
            low_bounds[id_start] = global_fit_params[id_start] - 1
            re_fit = True

        id_start += 1

    difference_to_upper = np.array([np.abs((a-b)/a) if a != np.inf and a != 0  else np.inf for a, b in zip(high_bounds[id_start:], global_fit_params[id_start:])])
    difference_to_lower = np.array([np.abs((a-b)/a) if b != -np.inf and a != 0 else np.inf for a, b in zip(global_fit_params[id_start:], low_bounds[id_start:])])

    # Evaluate all the other parameters
    for i in (range(len(global_fit_params)-id_start)):

        diff_to_high_i = difference_to_upper[i]
        diff_to_low_i = difference_to_lower[i]

        if diff_to_high_i < threshold:

            value = high_bounds[i+id_start]

            high_bounds[i+id_start] = value * 50 if value > 0 else value / 50
            re_fit = True

        if diff_to_low_i < threshold:

            value = low_bounds[i+id_start]
            low_bounds[i+id_start] = value * 50 if value < 0 else value / 50
            re_fit = True

    return re_fit, p0, low_bounds, high_bounds

def evaluate_fitting_and_refit(
        global_fit_params,
        cov,
        predicted,
        high_bounds,
        low_bounds,
        p0,
        fit_m_dep,
        limited_cp,
        limited_dh,
        limited_tm,
        fixed_cp,
        kwargs,
        fit_fx,
        n = 3,
        threshold=0.05,
        fit_m_value=True,):

    """
    Evaluate if the fitted parameters are too close to the fitting boundaries.
    If they are, re-fit with new expanded boundaries

    Parameters
    ----------
    global_fit_params: array-like
        fitted parameters
    cov: array-like
        covariance matrix of the fitted parameters
    predicted: list
        list of lists with the fitted values
    high_bounds: array-like
        upper bounds of the fitting parameters
    low_bounds: array-like
        lower bounds of the fitting parameters
    p0: array-like
        initial guess for the fitting parameters
    fit_m_dep: boolean
        if the m-dependence on temperature is fitted
    limited_cp: boolean
        if the cp bounds are user-defined
    limited_dh: boolean
        if the DH bounds are user-defined
    limited_tm: boolean
        if the Tm values are user-defined
    fixed_cp: boolean
        if the cp value is fixed
    kwargs: dict
        dictionary with the arguments for the fitting function
    fit_fx: callable
        function to perform the fitting
    n: int, optional
        number of times to re-fit
    threshold : float, optional
        Threshold to compare if the fitted parameters are too close to the boundaries
    fit_m_value : bool, optional
        Whether m0 (m-value) is fitted (not in oligomeric models)

    Returns
    -------
    global_fit_params: array-like
        fitted parameters
    cov: array-like
        covariance matrix of the fitted parameters
    predicted: list
        list of lists with the fitted values
    p0: array-like
        initial guess for the fitting parameters
    low_bounds: array-like
        lower bounds of the fitting parameters
    high_bounds: array-like
        higher bounds of the fitting parameters
    """

    for _ in range(n):

        re_fit, p0_new, low_bounds_new, high_bounds_new = evaluate_need_to_refit(
            global_fit_params,
            high_bounds,
            low_bounds,
            p0,
            fit_m1=fit_m_dep,
            check_cp=not limited_cp,
            check_dh=not limited_dh,
            check_tm=not limited_tm,
            fixed_cp=fixed_cp,
            threshold=threshold,
            fit_m0=fit_m_value,
        )

        if re_fit:

            p0, low_bounds, high_bounds = p0_new, low_bounds_new, high_bounds_new

            kwargs['initial_parameters'] = p0
            kwargs['low_bounds'] = low_bounds
            kwargs['high_bounds'] = high_bounds

            global_fit_params, cov, predicted = fit_fx(**kwargs)

        else:

            break

    return global_fit_params, cov, predicted, p0, low_bounds, high_bounds


"""
PyChemelt package for the analysis of chemical and thermal denaturation data
"""

from .main import Sample
from .monomer import Monomer
from .thermal_oligomer import ThermalOligomer

from .utils.math import (
    get_rss,
    shift_temperature,
    constant_baseline,
    linear_baseline,
    quadratic_baseline,
    exponential_baseline
)

from .utils.signals import (
    signal_two_state_tc_unfolding,
    signal_two_state_t_unfolding,
    map_two_state_model_to_signal_fx
)

from .utils.plotting import (
    plot_unfolding
)


from .utils.fitting import (
    fit_line_robust,
    fit_quadratic_robust,
    fit_exponential_robust
)


"""
Module for fitting of thermal unfolding curves.

Tests the fitting methods used by ThermalOligomer for their capabilities of fitting to simulated data
"""

import numpy as np

from pychemelt.utils.fitting import (
    fit_oligomer_unfolding_single_slopes,
    fit_oligomer_unfolding_shared_slopes_many_signals,
    fit_oligomer_unfolding_many_signals,
)

from pychemelt.utils.math import constant_baseline

from pychemelt.utils.signals import (
    map_two_state_model_to_signal_fx
)

# Centralized test constants
RNG_SEED = 2
TEMP_START = 30.0
TEMP_STOP = 90.0
N_TEMPS = 80
CONCS = np.arange(10, 100, 10)*1e-6

# Model / ground-truth parameters
DHm_VAL = 250
Tm_VAL = 70
CP0_VAL = 1.8


INTERCEPT_N = 50
C_N_VAL = 0
INTERCEPT_U = 100
C_U_VAL = 0

rng = np.random.default_rng(RNG_SEED)

def_params = {
    'dHm': DHm_VAL,
    'Tm': Tm_VAL+273.15,
    'Cp': CP0_VAL,
    'p1_N': C_N_VAL,
    'p2_N': INTERCEPT_N,
    'p3_N': 0,
    'p4_N': 0,
    'p1_U': C_U_VAL,
    'p2_U': INTERCEPT_U,
    'p3_U': 0,
    'p4_U': 0,
    'baseline_N_fx':constant_baseline,
    'baseline_U_fx':constant_baseline,
}

concs = CONCS

# Calculate signal range for proper y-axis scaling
temp_range  = np.linspace(TEMP_START, TEMP_STOP, N_TEMPS)
temp_range_K = temp_range + 273.15


def test_fit_monomer_unfolding_single_slopes_constant():
    signal_fx = map_two_state_model_to_signal_fx("Monomer")

    signal_list = []
    temp_list = []

    for C in concs:
        y = signal_fx(temp_range_K, C, **def_params)

        # Add gaussian error to signal
        y += rng.normal(0, 0.002*1e-3, len(y))

        signal_list.append(y)
        temp_list.append(temp_range)

    p0 = [Tm_VAL, DHm_VAL, CP0_VAL] + [INTERCEPT_N] *len(concs) + [INTERCEPT_U] * len(concs)
    low_bounds = [TEMP_START, TEMP_START, 1]   + [1e-5]*(2*len(concs))
    high_bounds = [TEMP_STOP, DHm_VAL+100, 5] + [1e3]*(2*len(concs))

    kwargs = {
        'list_of_temperatures' : temp_list,
        'list_of_signals' : signal_list,
        'oligomer_concentrations' : concs,
        'signal_fx' : signal_fx,
        'baseline_native_fx':constant_baseline,
        'baseline_unfolded_fx':constant_baseline, 
    }

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_single_slopes(
        initial_parameters=p0,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        **kwargs
    )

    expected = [Tm_VAL, DHm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:3], expected, rtol=0.1, atol=0)

    # Fit with fixed Tm
    p0_tm = p0.copy()
    low_bounds_tm = low_bounds.copy()
    high_bounds_tm = high_bounds.copy()

    p0_tm.pop(0)
    low_bounds_tm.pop(0)
    high_bounds_tm.pop(0)

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_single_slopes(
        initial_parameters=p0_tm,
        low_bounds=low_bounds_tm,
        high_bounds=high_bounds_tm,
        tm_value=Tm_VAL,
        **kwargs
    )

    expected = [DHm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:2], expected, rtol=0.1, atol=0)

    # End of - Fit with fixed Tm


    # Fit with fixed dH
    p0_dh = p0.copy()
    low_bounds_dh = low_bounds.copy()
    high_bounds_dh = high_bounds.copy()

    p0_dh.pop(1)
    low_bounds_dh.pop(1)
    high_bounds_dh.pop(1)

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_single_slopes(
        initial_parameters=p0_dh,
        low_bounds=low_bounds_dh,
        high_bounds=high_bounds_dh,
        dh_value=DHm_VAL,
        **kwargs
    )

    expected = [Tm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:2], expected, rtol=0.1, atol=0)

    # End of - Fit with fixed dH

    # Fit with fixed Cp
    p0_cp = p0.copy()
    low_bounds_cp = low_bounds.copy()
    high_bounds_cp = high_bounds.copy()

    p0_cp.pop(2)
    low_bounds_cp.pop(2)
    high_bounds_cp.pop(2)

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_single_slopes(
        initial_parameters=p0_cp,
        low_bounds=low_bounds_cp,
        high_bounds=high_bounds_cp,
        cp_value=CP0_VAL,
        **kwargs
    )

    expected = [Tm_VAL, DHm_VAL]

    np.testing.assert_allclose(global_fit_params[:2], expected, rtol=0.1, atol=0)

    # End of - Fit with fixed Cp

def test_fit_dimer_unfolding_single_slopes_constant():
    signal_fx = map_two_state_model_to_signal_fx("Dimer")

    signal_list = []
    temp_list = []

    for C in concs:
        y = signal_fx(temp_range_K, C, **def_params)

        # Add gaussian error to signal
        y += rng.normal(0, 0.002*1e-3, len(y))

        signal_list.append(y)
        temp_list.append(temp_range)

    p0 = [Tm_VAL, DHm_VAL, CP0_VAL] + [INTERCEPT_N] *len(concs) + [INTERCEPT_U] * len(concs)
    low_bounds = [TEMP_START, TEMP_START, 1] + [1e-5] * (2 * len(concs))
    high_bounds = [TEMP_STOP, DHm_VAL+100, 5] + [1e3] * (2 * len(concs))

    kwargs = {
        'list_of_temperatures': temp_list,
        'list_of_signals': signal_list,
        'oligomer_concentrations': concs,
        'signal_fx': signal_fx,
        'baseline_native_fx': constant_baseline,
        'baseline_unfolded_fx': constant_baseline, 
    }

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_single_slopes(
        initial_parameters=p0,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        **kwargs
    )

    expected = [Tm_VAL, DHm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:3], expected, rtol=0.1, atol=0)

    # Fit with fixed Tm
    p0_tm = p0.copy()
    low_bounds_tm = low_bounds.copy()
    high_bounds_tm = high_bounds.copy()

    p0_tm.pop(0)
    low_bounds_tm.pop(0)
    high_bounds_tm.pop(0)

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_single_slopes(
        initial_parameters=p0_tm,
        low_bounds=low_bounds_tm,
        high_bounds=high_bounds_tm,
        tm_value=Tm_VAL,
        **kwargs
    )

    expected = [DHm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:2], expected, rtol=0.1, atol=0)

    # End of - Fit with fixed Tm

    # Fit with fixed dH
    p0_dh = p0.copy()
    low_bounds_dh = low_bounds.copy()
    high_bounds_dh = high_bounds.copy()

    p0_dh.pop(1)
    low_bounds_dh.pop(1)
    high_bounds_dh.pop(1)

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_single_slopes(
        initial_parameters=p0_dh,
        low_bounds=low_bounds_dh,
        high_bounds=high_bounds_dh,
        dh_value=DHm_VAL,
        **kwargs
    )

    expected = [Tm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:2], expected, rtol=0.1, atol=0)

    # End of - Fit with fixed dH

    # Fit with fixed Cp
    p0_cp = p0.copy()
    low_bounds_cp = low_bounds.copy()
    high_bounds_cp = high_bounds.copy()

    p0_cp.pop(2)
    low_bounds_cp.pop(2)
    high_bounds_cp.pop(2)

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_single_slopes(
        initial_parameters=p0_cp,
        low_bounds=low_bounds_cp,
        high_bounds=high_bounds_cp,
        cp_value=CP0_VAL,
        **kwargs
    )

    expected = [Tm_VAL, DHm_VAL]

    np.testing.assert_allclose(global_fit_params[:2], expected, rtol=0.1, atol=0)

    # End of - Fit with fixed Cp


def test_fit_trimer_unfolding_single_slopes_constant():
    signal_fx = map_two_state_model_to_signal_fx("Trimer")

    signal_list = []
    temp_list = []

    for C in concs:
        y = signal_fx(temp_range_K, C, **def_params)

        # Add gaussian error to signal
        y += rng.normal(0, 0.002*1e-3, len(y))

        signal_list.append(y)
        temp_list.append(temp_range)

    p0 = [Tm_VAL, DHm_VAL, CP0_VAL] + [INTERCEPT_N] *len(concs) + [INTERCEPT_U] * len(concs)
    low_bounds = [TEMP_START, TEMP_START, 1] + [1e-5] * (2 * len(concs))
    high_bounds = [TEMP_STOP, DHm_VAL+100, 5] + [1e3] * (2 * len(concs))

    kwargs = {
        'list_of_temperatures': temp_list,
        'list_of_signals': signal_list,
        'oligomer_concentrations': concs,
        'signal_fx': signal_fx,
        'baseline_native_fx': constant_baseline,
        'baseline_unfolded_fx': constant_baseline, 
    }

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_single_slopes(
        initial_parameters=p0,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        **kwargs
    )

    expected = [Tm_VAL, DHm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:3], expected, rtol=0.1, atol=0)

    # Fit with fixed Tm
    p0_tm = p0.copy()
    low_bounds_tm = low_bounds.copy()
    high_bounds_tm = high_bounds.copy()

    p0_tm.pop(0)
    low_bounds_tm.pop(0)
    high_bounds_tm.pop(0)

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_single_slopes(
        initial_parameters=p0_tm,
        low_bounds=low_bounds_tm,
        high_bounds=high_bounds_tm,
        tm_value=Tm_VAL,
        **kwargs
    )

    expected = [DHm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:2], expected, rtol=0.1, atol=0)

    # End of - Fit with fixed Tm

    # Fit with fixed dH
    p0_dh = p0.copy()
    low_bounds_dh = low_bounds.copy()
    high_bounds_dh = high_bounds.copy()

    p0_dh.pop(1)
    low_bounds_dh.pop(1)
    high_bounds_dh.pop(1)

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_single_slopes(
        initial_parameters=p0_dh,
        low_bounds=low_bounds_dh,
        high_bounds=high_bounds_dh,
        dh_value=DHm_VAL,
        **kwargs
    )

    expected = [Tm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:2], expected, rtol=0.1, atol=0)

    # End of - Fit with fixed dH

    # Fit with fixed Cp
    p0_cp = p0.copy()
    low_bounds_cp = low_bounds.copy()
    high_bounds_cp = high_bounds.copy()

    p0_cp.pop(2)
    low_bounds_cp.pop(2)
    high_bounds_cp.pop(2)

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_single_slopes(
        initial_parameters=p0_cp,
        low_bounds=low_bounds_cp,
        high_bounds=high_bounds_cp,
        cp_value=CP0_VAL,
        **kwargs
    )

    expected = [Tm_VAL, DHm_VAL]

    np.testing.assert_allclose(global_fit_params[:2], expected, rtol=0.1, atol=0)

    # End of - Fit with fixed Cp

def test_fit_tetramer_unfolding_single_slopes_constant():
    signal_fx = map_two_state_model_to_signal_fx("Tetramer")

    signal_list = []
    temp_list = []

    for C in concs:
        y = signal_fx(temp_range_K, C, **def_params)

        # Add gaussian error to signal
        y += rng.normal(0, 0.002*1e-3, len(y))

        signal_list.append(y)
        temp_list.append(temp_range)

    p0 = [Tm_VAL, DHm_VAL, CP0_VAL] + [INTERCEPT_N] *len(concs) + [INTERCEPT_U] * len(concs)
    low_bounds = [TEMP_START, TEMP_START, 1] + [1e-5] * (2 * len(concs))
    high_bounds = [TEMP_STOP, DHm_VAL+100, 5] + [1e3] * (2 * len(concs))

    kwargs = {
        'list_of_temperatures': temp_list,
        'list_of_signals': signal_list,
        'oligomer_concentrations': concs,
        'signal_fx': signal_fx,
        'baseline_native_fx': constant_baseline,
        'baseline_unfolded_fx': constant_baseline, 
    }

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_single_slopes(
        initial_parameters=p0,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        **kwargs
    )

    expected = [Tm_VAL, DHm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:3], expected, rtol=0.1, atol=0)

    # Fit with fixed Tm
    p0_tm = p0.copy()
    low_bounds_tm = low_bounds.copy()
    high_bounds_tm = high_bounds.copy()

    p0_tm.pop(0)
    low_bounds_tm.pop(0)
    high_bounds_tm.pop(0)

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_single_slopes(
        initial_parameters=p0_tm,
        low_bounds=low_bounds_tm,
        high_bounds=high_bounds_tm,
        tm_value=Tm_VAL,
        **kwargs
    )

    expected = [DHm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:2], expected, rtol=0.1, atol=0)

    # End of - Fit with fixed Tm

    # Fit with fixed dH
    p0_dh = p0.copy()
    low_bounds_dh = low_bounds.copy()
    high_bounds_dh = high_bounds.copy()

    p0_dh.pop(1)
    low_bounds_dh.pop(1)
    high_bounds_dh.pop(1)

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_single_slopes(
        initial_parameters=p0_dh,
        low_bounds=low_bounds_dh,
        high_bounds=high_bounds_dh,
        dh_value=DHm_VAL,
        **kwargs
    )

    expected = [Tm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:2], expected, rtol=0.1, atol=0)

    # End of - Fit with fixed dH

    # Fit with fixed Cp
    p0_cp = p0.copy()
    low_bounds_cp = low_bounds.copy()
    high_bounds_cp = high_bounds.copy()

    p0_cp.pop(2)
    low_bounds_cp.pop(2)
    high_bounds_cp.pop(2)

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_single_slopes(
        initial_parameters=p0_cp,
        low_bounds=low_bounds_cp,
        high_bounds=high_bounds_cp,
        cp_value=CP0_VAL,
        **kwargs
    )

    expected = [Tm_VAL, DHm_VAL]

    np.testing.assert_allclose(global_fit_params[:2], expected, rtol=0.1, atol=0)

    # End of - Fit with fixed Cp


# Test fitting global slope

def test_fit_monomer_unfolding_shared_slopes_many_signals_constant():
    signal_fx = map_two_state_model_to_signal_fx("Monomer")

    signal_list = []
    temp_list = []

    for C in concs:
        y = signal_fx(temp_range_K, C, **def_params)

        # Add gaussian error to signal
        y += rng.normal(0, 0.002*1e-3, len(y))

        signal_list.append(y)
        temp_list.append(temp_range)

    p0 = [Tm_VAL, DHm_VAL, CP0_VAL] + [INTERCEPT_N] *len(concs) + [INTERCEPT_U] * len(concs)
    low_bounds = [TEMP_START, TEMP_START, 1]   + [1e-5]*(2*len(concs))
    high_bounds = [TEMP_STOP, DHm_VAL+100, 5] + [1e3]*(2*len(concs))

    kwargs = {
        'list_of_temperatures' : temp_list,
        'list_of_signals' : signal_list,
        'signal_ids': [0 for _ in range(len(signal_list))],
        'oligomer_concentrations' : concs,
        'signal_fx' : signal_fx,
        'baseline_native_fx':constant_baseline,
        'baseline_unfolded_fx':constant_baseline, 
    }

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_shared_slopes_many_signals(
        initial_parameters=p0,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        **kwargs
    )

    expected = [Tm_VAL, DHm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:3], expected, rtol=0.1, atol=0)

    # Fit with fixed Tm
    p0_tm = p0.copy()
    low_bounds_tm = low_bounds.copy()
    high_bounds_tm = high_bounds.copy()

    p0_tm.pop(0)
    low_bounds_tm.pop(0)
    high_bounds_tm.pop(0)

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_shared_slopes_many_signals(
        initial_parameters=p0_tm,
        low_bounds=low_bounds_tm,
        high_bounds=high_bounds_tm,
        tm_value=Tm_VAL,
        **kwargs
    )

    expected = [DHm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:2], expected, rtol=0.1, atol=0)

    # End of - Fit with fixed Tm

def test_fit_dimer_unfolding_shared_slopes_many_signals_constant():
    signal_fx = map_two_state_model_to_signal_fx("Dimer")

    signal_list = []
    temp_list = []

    for C in concs:
        y = signal_fx(temp_range_K, C, **def_params)

        # Add gaussian error to signal
        y += rng.normal(0, 0.002*1e-3, len(y))

        signal_list.append(y)
        temp_list.append(temp_range)

    p0 = [Tm_VAL, DHm_VAL, CP0_VAL] + [INTERCEPT_N] *len(concs) + [INTERCEPT_U] * len(concs)
    low_bounds = [TEMP_START, TEMP_START, 1] + [1e-5] * (2 * len(concs))
    high_bounds = [TEMP_STOP, DHm_VAL+100, 5] + [1e3] * (2 * len(concs))

    kwargs = {
        'list_of_temperatures': temp_list,
        'list_of_signals': signal_list,
        'signal_ids': [0 for _ in range(len(signal_list))],
        'oligomer_concentrations': concs,
        'signal_fx': signal_fx,
        'baseline_native_fx': constant_baseline,
        'baseline_unfolded_fx': constant_baseline, 
    }

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_shared_slopes_many_signals(
        initial_parameters=p0,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        **kwargs
    )

    expected = [Tm_VAL, DHm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:3], expected, rtol=0.1, atol=0)

    # Fit with fixed Tm
    p0_tm = p0.copy()
    low_bounds_tm = low_bounds.copy()
    high_bounds_tm = high_bounds.copy()

    p0_tm.pop(0)
    low_bounds_tm.pop(0)
    high_bounds_tm.pop(0)

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_shared_slopes_many_signals(
        initial_parameters=p0_tm,
        low_bounds=low_bounds_tm,
        high_bounds=high_bounds_tm,
        tm_value=Tm_VAL,
        **kwargs
    )

    expected = [DHm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:2], expected, rtol=0.1, atol=0)

    # End of - Fit with fixed Tm


def test_fit_trimer_unfolding_shared_slopes_many_signals_constant():
    signal_fx = map_two_state_model_to_signal_fx("Trimer")

    signal_list = []
    temp_list = []

    for C in concs:
        y = signal_fx(temp_range_K, C, **def_params)

        # Add gaussian error to signal
        y += rng.normal(0, 0.002*1e-3, len(y))

        signal_list.append(y)
        temp_list.append(temp_range)

    p0 = [Tm_VAL, DHm_VAL, CP0_VAL] + [INTERCEPT_N] *len(concs) + [INTERCEPT_U] * len(concs)
    low_bounds = [TEMP_START, TEMP_START, 1] + [1e-5] * (2 * len(concs))
    high_bounds = [TEMP_STOP, DHm_VAL+100, 5] + [1e3] * (2 * len(concs))

    kwargs = {
        'list_of_temperatures': temp_list,
        'list_of_signals': signal_list,
        'signal_ids': [0 for _ in range(len(signal_list))],
        'oligomer_concentrations': concs,
        'signal_fx': signal_fx,
        'baseline_native_fx': constant_baseline,
        'baseline_unfolded_fx': constant_baseline, 
    }

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_shared_slopes_many_signals(
        initial_parameters=p0,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        **kwargs
    )

    expected = [Tm_VAL, DHm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:3], expected, rtol=0.1, atol=0)

    # Fit with fixed Tm
    p0_tm = p0.copy()
    low_bounds_tm = low_bounds.copy()
    high_bounds_tm = high_bounds.copy()

    p0_tm.pop(0)
    low_bounds_tm.pop(0)
    high_bounds_tm.pop(0)

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_shared_slopes_many_signals(
        initial_parameters=p0_tm,
        low_bounds=low_bounds_tm,
        high_bounds=high_bounds_tm,
        tm_value=Tm_VAL,
        **kwargs
    )

    expected = [DHm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:2], expected, rtol=0.1, atol=0)

    # End of - Fit with fixed Tm


def test_fit_tetramer_unfolding_shared_slopes_many_signals_constant():
    signal_fx = map_two_state_model_to_signal_fx("Tetramer")

    signal_list = []
    temp_list = []

    for C in concs:
        y = signal_fx(temp_range_K, C, **def_params)

        # Add gaussian error to signal
        y += rng.normal(0, 0.002*1e-3, len(y))

        signal_list.append(y)
        temp_list.append(temp_range)

    p0 = [Tm_VAL, DHm_VAL, CP0_VAL] + [INTERCEPT_N] *len(concs) + [INTERCEPT_U] * len(concs)
    low_bounds = [TEMP_START, TEMP_START, 1] + [1e-5] * (2 * len(concs))
    high_bounds = [TEMP_STOP, DHm_VAL+100, 5] + [1e3] * (2 * len(concs))

    kwargs = {
        'list_of_temperatures': temp_list,
        'list_of_signals': signal_list,
        'signal_ids': [0 for _ in range(len(signal_list))],
        'oligomer_concentrations': concs,
        'signal_fx': signal_fx,
        'baseline_native_fx': constant_baseline,
        'baseline_unfolded_fx': constant_baseline, 
    }

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_shared_slopes_many_signals(
        initial_parameters=p0,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        **kwargs
    )

    expected = [Tm_VAL, DHm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:3], expected, rtol=0.1, atol=0)

    # Fit with fixed Tm
    p0_tm = p0.copy()
    low_bounds_tm = low_bounds.copy()
    high_bounds_tm = high_bounds.copy()

    p0_tm.pop(0)
    low_bounds_tm.pop(0)
    high_bounds_tm.pop(0)

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_shared_slopes_many_signals(
        initial_parameters=p0_tm,
        low_bounds=low_bounds_tm,
        high_bounds=high_bounds_tm,
        tm_value=Tm_VAL,
        **kwargs
    )

    expected = [DHm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:2], expected, rtol=0.1, atol=0)

    # End of - Fit with fixed Tm


# Test fitting global slope and baselines

def test_fit_monomer_unfolding_many_signals_constant():
    signal_fx = map_two_state_model_to_signal_fx("Monomer")

    signal_list = []
    temp_list = []

    for C in concs:
        y = signal_fx(temp_range_K, C, **def_params)

        # Add gaussian error to signal
        y += rng.normal(0, 0.002*1e-3, len(y))

        signal_list.append(y)
        temp_list.append(temp_range)

    p0 = [Tm_VAL, DHm_VAL, CP0_VAL] + [INTERCEPT_N, INTERCEPT_U] + [0] * (2 * 3)
    low_bounds = [TEMP_START, TEMP_START, 1] + [-1e2] * (2 * 4)
    high_bounds = [TEMP_STOP, DHm_VAL+100, 5] + [1e3] * (2 * 4)

    kwargs = {
        'list_of_temperatures': temp_list,
        'list_of_signals': signal_list,
        'signal_ids': [0 for _ in range(len(signal_list))],
        'oligomer_concentrations': concs,
        'signal_fx': signal_fx,
        'baseline_native_fx': constant_baseline,
        'baseline_unfolded_fx': constant_baseline, 
        'fit_native_olig_slope' : False,
        'fit_unfolded_olig_slope' : False,
    }

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_many_signals(
        initial_parameters=p0,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        **kwargs
    )

    expected = [Tm_VAL, DHm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:3], expected, rtol=0.1, atol=0)

    # Fit scale factor

    scale_factors = [1 for _ in range(len(signal_list) - 1)]
    scale_factors_low = [0.5882 for _ in range(len(signal_list) - 1)]
    scale_factors_high = [1.7 for _ in range(len(signal_list) - 1)]

    p0 = np.concatenate([p0, scale_factors])
    low_bounds = np.concatenate([low_bounds, scale_factors_low])
    high_bounds = np.concatenate([high_bounds, scale_factors_high])

    scale_factor_exclude_ids = [len(signal_list) - 1]

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_many_signals(
        initial_parameters=p0,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        model_scale_factor=True,
        scale_factor_exclude_ids = scale_factor_exclude_ids,
        **kwargs
    )

    expected = [Tm_VAL, DHm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:3], expected, rtol=0.1, atol=0)

    # Fit cp value set

    p0 = [Tm_VAL, DHm_VAL] + [INTERCEPT_N, INTERCEPT_U] + [0] * (2 * 3)
    low_bounds = [TEMP_START, TEMP_START] + [-1e2] * (2 * 4)
    high_bounds = [TEMP_STOP, DHm_VAL+100] + [1e3] * (2 * 4)

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_many_signals(
        initial_parameters=p0,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        cp_value=CP0_VAL,
        **kwargs
    )

    expected = [Tm_VAL, DHm_VAL]

    np.testing.assert_allclose(global_fit_params[:2], expected, rtol=0.1, atol=0)

def test_fit_dimer_unfolding_many_signals_constant():
    signal_fx = map_two_state_model_to_signal_fx("Dimer")

    signal_list = []
    temp_list = []

    for C in concs:
        y = signal_fx(temp_range_K, C, **def_params)

        # Add gaussian error to signal
        y += rng.normal(0, 0.002*1e-3, len(y))

        signal_list.append(y)
        temp_list.append(temp_range)

    p0 = [Tm_VAL, DHm_VAL, CP0_VAL] + [INTERCEPT_N, INTERCEPT_U] + [0] * (2 * 3)
    low_bounds = [TEMP_START, TEMP_START, 1] + [-1e2] * (2 * 4)
    high_bounds = [TEMP_STOP, DHm_VAL+100, 5] + [1e3] * (2 * 4)

    kwargs = {
        'list_of_temperatures': temp_list,
        'list_of_signals': signal_list,
        'signal_ids': [0 for _ in range(len(signal_list))],
        'oligomer_concentrations': concs,
        'signal_fx': signal_fx,
        'baseline_native_fx': constant_baseline,
        'baseline_unfolded_fx': constant_baseline, 
        'fit_native_olig_slope': False,
        'fit_unfolded_olig_slope': False,
    }

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_many_signals(
        initial_parameters=p0,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        **kwargs
    )

    expected = [Tm_VAL, DHm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:3], expected, rtol=0.1, atol=0)

    # Fit scale factor

    scale_factors = [1 for _ in range(len(signal_list) - 1)]
    scale_factors_low = [0.5882 for _ in range(len(signal_list) - 1)]
    scale_factors_high = [1.7 for _ in range(len(signal_list) - 1)]

    p0 = np.concatenate([p0, scale_factors])
    low_bounds = np.concatenate([low_bounds, scale_factors_low])
    high_bounds = np.concatenate([high_bounds, scale_factors_high])

    scale_factor_exclude_ids = [len(signal_list) - 1]

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_many_signals(
        initial_parameters=p0,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        model_scale_factor=True,
        scale_factor_exclude_ids=scale_factor_exclude_ids,
        **kwargs
    )

    expected = [Tm_VAL, DHm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:3], expected, rtol=0.1, atol=0)

    # Fit cp value set

    p0 = [Tm_VAL, DHm_VAL] + [INTERCEPT_N, INTERCEPT_U] + [0] * (2 * 3)
    low_bounds = [TEMP_START, TEMP_START] + [-1e2] * (2 * 4)
    high_bounds = [TEMP_STOP, DHm_VAL+100] + [1e3] * (2 * 4)

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_many_signals(
        initial_parameters=p0,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        cp_value=CP0_VAL,
        **kwargs
    )

    expected = [Tm_VAL, DHm_VAL]

    np.testing.assert_allclose(global_fit_params[:2], expected, rtol=0.1, atol=0)

def test_fit_trimer_unfolding_many_signals_constant():
    signal_fx = map_two_state_model_to_signal_fx("Trimer")

    signal_list = []
    temp_list = []

    for C in concs:
        y = signal_fx(temp_range_K, C, **def_params)

        # Add gaussian error to signal
        y += rng.normal(0, 0.002*1e-3, len(y))

        signal_list.append(y)
        temp_list.append(temp_range)

    p0 = [Tm_VAL, DHm_VAL, CP0_VAL] + [INTERCEPT_N, INTERCEPT_U] + [0] * (2 * 3)
    low_bounds = [TEMP_START, TEMP_START, 1] + [-1e2] * (2 * 4)
    high_bounds = [TEMP_STOP, DHm_VAL+100, 5] + [1e3] * (2 * 4)

    kwargs = {
        'list_of_temperatures': temp_list,
        'list_of_signals': signal_list,
        'signal_ids': [0 for _ in range(len(signal_list))],
        'oligomer_concentrations': concs,
        'signal_fx': signal_fx,
        'baseline_native_fx': constant_baseline,
        'baseline_unfolded_fx': constant_baseline, 
        'fit_native_olig_slope': False,
        'fit_unfolded_olig_slope': False,
    }

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_many_signals(
        initial_parameters=p0,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        **kwargs
    )

    expected = [Tm_VAL, DHm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:3], expected, rtol=0.1, atol=0)

    # Fit scale factor

    scale_factors = [1 for _ in range(len(signal_list) - 1)]
    scale_factors_low = [0.5882 for _ in range(len(signal_list) - 1)]
    scale_factors_high = [1.7 for _ in range(len(signal_list) - 1)]

    p0 = np.concatenate([p0, scale_factors])
    low_bounds = np.concatenate([low_bounds, scale_factors_low])
    high_bounds = np.concatenate([high_bounds, scale_factors_high])

    scale_factor_exclude_ids = [len(signal_list) - 1]

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_many_signals(
        initial_parameters=p0,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        model_scale_factor=True,
        scale_factor_exclude_ids=scale_factor_exclude_ids,
        **kwargs
    )

    expected = [Tm_VAL, DHm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:3], expected, rtol=0.1, atol=0)

    # Fit cp value set

    p0 = [Tm_VAL, DHm_VAL] + [INTERCEPT_N, INTERCEPT_U] + [0] * (2 * 3)
    low_bounds = [TEMP_START, TEMP_START] + [-1e2] * (2 * 4)
    high_bounds = [TEMP_STOP, DHm_VAL+100] + [1e3] * (2 * 4)

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_many_signals(
        initial_parameters=p0,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        cp_value=CP0_VAL,
        **kwargs
    )

    expected = [Tm_VAL, DHm_VAL]

    np.testing.assert_allclose(global_fit_params[:2], expected, rtol=0.1, atol=0)

def test_fit_tetramer_unfolding_many_signals_constant():
    signal_fx = map_two_state_model_to_signal_fx("Tetramer")

    signal_list = []
    temp_list = []

    for C in concs:
        y = signal_fx(temp_range_K, C, **def_params)
        # Add gaussian error to signal
        y += rng.normal(0, 0.002*1e-3, len(y))

        signal_list.append(y)
        temp_list.append(temp_range)

    p0 = [Tm_VAL, DHm_VAL, CP0_VAL] + [INTERCEPT_N, INTERCEPT_U] + [0] * (2 * 3)
    low_bounds = [TEMP_START, TEMP_START, 1] + [-1e2] * (2 * 4)
    high_bounds = [TEMP_STOP, DHm_VAL+100, 5] + [1e3] * (2 * 4)

    kwargs = {
        'list_of_temperatures': temp_list,
        'list_of_signals': signal_list,
        'signal_ids': [0 for _ in range(len(signal_list))],
        'oligomer_concentrations': concs,
        'signal_fx': signal_fx,
        'baseline_native_fx': constant_baseline,
        'baseline_unfolded_fx': constant_baseline, 
        'fit_native_olig_slope': False,
        'fit_unfolded_olig_slope': False,
    }

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_many_signals(
        initial_parameters=p0,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        **kwargs
    )

    expected = [Tm_VAL, DHm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:3], expected, rtol=0.1, atol=0)

    # Fit scale factor

    scale_factors = [1 for _ in range(len(signal_list) - 1)]
    scale_factors_low = [0.5882 for _ in range(len(signal_list) - 1)]
    scale_factors_high = [1.7 for _ in range(len(signal_list) - 1)]

    p0 = np.concatenate([p0, scale_factors])
    low_bounds = np.concatenate([low_bounds, scale_factors_low])
    high_bounds = np.concatenate([high_bounds, scale_factors_high])

    scale_factor_exclude_ids = [len(signal_list) - 1]

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_many_signals(
        initial_parameters=p0,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        model_scale_factor=True,
        scale_factor_exclude_ids=scale_factor_exclude_ids,
        **kwargs
    )

    expected = [Tm_VAL, DHm_VAL, CP0_VAL]

    np.testing.assert_allclose(global_fit_params[:3], expected, rtol=0.1, atol=0)

    # Fit cp value set

    p0 = [Tm_VAL, DHm_VAL] + [INTERCEPT_N, INTERCEPT_U] + [0] * (2 * 3)
    low_bounds = [TEMP_START, TEMP_START] + [-1e2] * (2 * 4)
    high_bounds = [TEMP_STOP, DHm_VAL+100] + [1e3] * (2 * 4)

    global_fit_params, cov, predicted_lst = fit_oligomer_unfolding_many_signals(
        initial_parameters=p0,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        cp_value=CP0_VAL,
        **kwargs
    )

    expected = [Tm_VAL, DHm_VAL]

    np.testing.assert_allclose(global_fit_params[:2], expected, rtol=0.1, atol=0)


"""
This module contains helper functions to obtain equilibrium constants
Author: Osvaldo Burastero

Useful references for unfolding models:
    - Rumfeldt, Jessica AO, et al. "Conformational stability and folding mechanisms of dimeric proteins." Progress in biophysics and molecular biology 98.1 (2008): 61-84.
    - Bedouelle, Hugues. "Principles and equations for measuring and interpreting protein stability: From monomer to tetramer." Biochimie 121 (2016): 29-37.
    - Mazurenko, Stanislav, et al. "Exploration of protein unfolding by modelling calorimetry data from reheating." Scientific reports 7.1 (2017): 16321.

All thermodynamic parameters are used in kcal mol units

Unfolding functions for monomers have an argument called 'extra_arg' that is not used.
This is because unfolding functions for oligomers require the protein concentration in that position

"""

import numpy as np

from .math import shift_temperature_K

from .constants import R_gas

def eq_constant_thermo(T,DH1,T1,Cp):

    """
    T1 is the temperature at which ΔG(T) = 0
    ΔH1, the variation of enthalpy between the two considered states at T1
    Cp the variation of calorific capacity between the two states

    Parameters
    ----------
    T : array-like
        Temperature (Kelvin)
    DH1 : float
        Variation of enthalpy between the two considered states at T1 (kcal/mol)
    T1 : float
        Temperature at which the equilibrium constant equals one (Kelvin)
    Cp : float
        Variation of heat capacity between the two states (kcal/mol/K)

    Returns
    -------
    numpy.ndarray
        Equilibrium constant at the given temperature
    """

    DG = DH1*(1 - T/T1) - Cp*(T1 - T + T*np.log(T/T1))
    K  = np.exp(-DG / (R_gas * T))

    return K


def eq_constant_termochem(T,D,DHm,Tm,Cp0,m0,m1):

    """
    Ref: Louise Hamborg et al., 2020. Global analysis of protein stability by temperature and chemical
    denaturation

    Parameters
    ----------
    T : array-like
        Temperature (Kelvin only!)
    D : float
        Denaturant concentration (M)
    DHm : float
        Enthalpy change at Tm (kcal/mol)
    Tm : float
        Melting temperature where ΔG = 0 (Kelvin only!)
    Cp0 : float
        Heat capacity change (kcal/mol/K)
    m0 : float
        m-value at the reference temperature
    m1 : float
        Temperature dependence of the m-value

    Returns
    -------
    numpy.ndarray
        Equilibrium constant at a certain temperature and denaturant agent concentration
    """

    DT  = shift_temperature_K(T)

    DG   = DHm*(1 - T/Tm) + Cp0*(T - Tm - T*np.log(T/Tm)) - D*(m0 + m1*DT)

    DG_RT = -DG / (R_gas * T)

    K     = np.exp(DG_RT)

    return K